from dataclasses import dataclass
from datetime import datetime, timezone

from dataclass_csv import accept_whitespaces, dateformat
from sqlalchemy import Column, DateTime, Float, ForeignKey, Integer, String
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship

import vigorish.database as db
from vigorish.util.datetime_util import make_tzaware, TIME_ZONE_NEW_YORK
from vigorish.util.dt_format_strings import CSV_UTC, DT_AWARE


class PitchFx(db.Base):

    __tablename__ = "pitchfx"
    id = Column(Integer, primary_key=True)
    bb_game_id = Column(String)
    bbref_game_id = Column(String)
    pitch_app_id = Column(String)
    inning_id = Column(String)
    at_bat_id = Column(String)
    pitcher_id_mlb = Column(Integer)
    batter_id_mlb = Column(Integer)
    pitcher_id_bbref = Column(String)
    batter_id_bbref = Column(String)
    pitcher_team_id_bb = Column(String)
    opponent_team_id_bb = Column(String)
    p_throws = Column(String)
    stand = Column(String)
    pitch_id = Column(Integer)
    inning = Column(Integer)
    ab_total = Column(Integer)
    ab_count = Column(Integer)
    ab_id = Column(Integer)
    table_row_number = Column(Integer)
    des = Column(String)
    strikes = Column(Integer)
    balls = Column(Integer)
    basic_type = Column(String)
    pdes = Column(String)
    mlbam_pitch_name = Column(String)
    start_speed = Column(Float)
    spin = Column(Float)
    zone_location = Column(Integer)
    sz_top = Column(Float)
    sz_bot = Column(Float)
    pfx_xdatafile = Column(Float)
    pfx_zdatafile = Column(Float)
    pfx_x = Column(Float)
    pfx_z = Column(Float)
    uncorrected_pfx_x = Column(Float)
    uncorrected_pfx_z = Column(Float)
    px = Column(Float)
    pz = Column(Float)
    pxold = Column(Float)
    pzold = Column(Float)
    park_sv_id = Column(String)
    game_start_time_utc = Column(DateTime)
    time_pitch_thrown_utc = Column(DateTime)
    seconds_since_game_start = Column(Integer)
    has_zone_location = Column(Integer)
    batter_did_swing = Column(Integer)
    batter_made_contact = Column(Integer)
    called_strike = Column(Integer)
    swinging_strike = Column(Integer)
    inside_strike_zone = Column(Integer)
    outside_strike_zone = Column(Integer)
    swing_inside_zone = Column(Integer)
    swing_outside_zone = Column(Integer)
    contact_inside_zone = Column(Integer)
    contact_outside_zone = Column(Integer)
    is_batted_ball = Column(Integer)
    is_ground_ball = Column(Integer)
    is_fly_ball = Column(Integer)
    is_line_drive = Column(Integer)
    is_pop_up = Column(Integer)
    is_final_pitch_of_ab = Column(Integer)
    ab_result_out = Column(Integer)
    ab_result_hit = Column(Integer)
    ab_result_single = Column(Integer)
    ab_result_double = Column(Integer)
    ab_result_triple = Column(Integer)
    ab_result_homerun = Column(Integer)
    ab_result_bb = Column(Integer)
    ab_result_ibb = Column(Integer)
    ab_result_k = Column(Integer)
    ab_result_hbp = Column(Integer)
    ab_result_error = Column(Integer)
    ab_result_sac_hit = Column(Integer)
    ab_result_sac_fly = Column(Integer)
    ab_result_unclear = Column(Integer)
    pitch_type_int = Column(Integer)
    pbp_play_result = Column(String)
    pbp_runs_outs_result = Column(String)
    is_sp = Column(Integer)
    is_rp = Column(Integer)
    is_patched = Column(Integer)
    is_invalid_ibb = Column(Integer)
    is_out_of_sequence = Column(Integer)

    pitcher_id = Column(Integer, ForeignKey("player.id"), index=True)
    batter_id = Column(Integer, ForeignKey("player.id"), index=True)
    team_pitching_id = Column(Integer, ForeignKey("team.id"))
    team_batting_id = Column(Integer, ForeignKey("team.id"))
    game_status_id = Column(Integer, ForeignKey("scrape_status_game.id"))
    pitch_app_db_id = Column(Integer, ForeignKey("scrape_status_pitch_app.id"), index=True)
    date_id = Column(Integer, ForeignKey("scrape_status_date.id"))
    season_id = Column(Integer, ForeignKey("season.id"))

    pitcher = relationship("Player", foreign_keys=[pitcher_id], backref="pfx_as_pitcher")
    batter = relationship("Player", foreign_keys=[batter_id], backref="pfx_as_batter")
    team_pitching = relationship("Team", foreign_keys=[team_pitching_id])
    team_batting = relationship("Team", foreign_keys=[team_batting_id])
    game = relationship("GameScrapeStatus", foreign_keys=[game_status_id])
    pitch_app = relationship("PitchAppScrapeStatus", foreign_keys=[pitch_app_db_id])
    date = relationship("DateScrapeStatus", foreign_keys=[date_id])
    season = relationship("Season", foreign_keys=[season_id])

    @hybrid_property
    def game_date(self):
        year = self.game_start_time.year
        month = self.game_start_time.month
        day = self.game_start_time.day
        return datetime(year, month, day)

    @hybrid_property
    def game_start_time(self):
        game_start_utc = make_tzaware(self.game_start_time_utc, use_tz=timezone.utc, localize=False)
        return make_tzaware(game_start_utc, use_tz=TIME_ZONE_NEW_YORK, localize=True)

    @hybrid_property
    def time_pitch_thrown(self):
        thrown_utc = make_tzaware(self.time_pitch_thrown_utc, use_tz=timezone.utc, localize=False)
        return make_tzaware(thrown_utc, use_tz=TIME_ZONE_NEW_YORK, localize=True)

    @classmethod
    def from_dict(cls, pfx_dict):
        pfx_dict = cls.update_pfx_dict(pfx_dict)
        pfx_dict["zone_location"] = int(pfx_dict["zone_location"])
        pfx_dict["batter_did_swing"] = int(pfx_dict["batter_did_swing"])
        pfx_dict["batter_made_contact"] = int(pfx_dict["batter_made_contact"])
        pfx_dict["called_strike"] = int(pfx_dict["called_strike"])
        pfx_dict["swinging_strike"] = int(pfx_dict["swinging_strike"])
        pfx_dict["inside_strike_zone"] = int(pfx_dict["inside_strike_zone"])
        pfx_dict["outside_strike_zone"] = int(pfx_dict["outside_strike_zone"])
        pfx_dict["swing_inside_zone"] = int(pfx_dict["swing_inside_zone"])
        pfx_dict["swing_outside_zone"] = int(pfx_dict["swing_outside_zone"])
        pfx_dict["contact_inside_zone"] = int(pfx_dict["contact_inside_zone"])
        pfx_dict["contact_outside_zone"] = int(pfx_dict["contact_outside_zone"])
        pfx_dict["is_batted_ball"] = int(pfx_dict["is_batted_ball"])
        pfx_dict["is_ground_ball"] = int(pfx_dict["is_ground_ball"])
        pfx_dict["is_fly_ball"] = int(pfx_dict["is_fly_ball"])
        pfx_dict["is_line_drive"] = int(pfx_dict["is_line_drive"])
        pfx_dict["is_pop_up"] = int(pfx_dict["is_pop_up"])
        pfx_dict["is_final_pitch_of_ab"] = int(pfx_dict["is_final_pitch_of_ab"])
        pfx_dict["ab_result_out"] = int(pfx_dict["ab_result_out"])
        pfx_dict["ab_result_hit"] = int(pfx_dict["ab_result_hit"])
        pfx_dict["ab_result_single"] = int(pfx_dict["ab_result_single"])
        pfx_dict["ab_result_double"] = int(pfx_dict["ab_result_double"])
        pfx_dict["ab_result_triple"] = int(pfx_dict["ab_result_triple"])
        pfx_dict["ab_result_homerun"] = int(pfx_dict["ab_result_homerun"])
        pfx_dict["ab_result_bb"] = int(pfx_dict["ab_result_bb"])
        pfx_dict["ab_result_ibb"] = int(pfx_dict["ab_result_ibb"])
        pfx_dict["ab_result_k"] = int(pfx_dict["ab_result_k"])
        pfx_dict["ab_result_hbp"] = int(pfx_dict["ab_result_hbp"])
        pfx_dict["ab_result_error"] = int(pfx_dict["ab_result_error"])
        pfx_dict["ab_result_sac_hit"] = int(pfx_dict["ab_result_sac_hit"])
        pfx_dict["ab_result_sac_fly"] = int(pfx_dict["ab_result_sac_fly"])
        pfx_dict["ab_result_unclear"] = int(pfx_dict["ab_result_unclear"])
        pfx_dict["is_sp"] = int(pfx_dict["is_sp"])
        pfx_dict["is_rp"] = int(pfx_dict["is_rp"])
        pfx_dict["spin"] = round(pfx_dict["spin"], 1)
        pfx_dict["pfx_x"] = round(pfx_dict["pfx_x"], 2)
        pfx_dict["pfx_z"] = round(pfx_dict["pfx_z"], 2)
        pfx_dict["px"] = round(pfx_dict["px"], 2)
        pfx_dict["pz"] = round(pfx_dict["pz"], 2)
        return cls(**pfx_dict)

    @staticmethod
    def update_pfx_dict(pfx_dict):
        game_start_str = pfx_dict.pop("game_start_time_str")
        pitch_thrown_str = pfx_dict.pop("time_pitch_thrown_str")
        game_start_time = datetime.strptime(game_start_str, DT_AWARE).astimezone(timezone.utc)
        time_pitch_thrown = (
            datetime.strptime(pitch_thrown_str, DT_AWARE).astimezone(timezone.utc) if pitch_thrown_str else None
        )
        seconds_since_game_start = (
            int((time_pitch_thrown - game_start_time).total_seconds()) if time_pitch_thrown else 0
        )
        pfx_dict["game_start_time_utc"] = game_start_time
        pfx_dict["time_pitch_thrown_utc"] = time_pitch_thrown
        pfx_dict["seconds_since_game_start"] = seconds_since_game_start
        pfx_dict["basic_type"] = pfx_dict.pop("type")
        pfx_dict["pitch_id"] = pfx_dict.pop("id")
        pfx_dict["pitcher_id_mlb"] = int(pfx_dict.pop("pitcher_id"))
        pfx_dict["batter_id_mlb"] = int(pfx_dict.pop("batter_id"))
        pfx_dict["pitcher_id"] = pfx_dict.pop("pitcher_id_db")
        pfx_dict["batter_id"] = pfx_dict.pop("batter_id_db")
        pfx_dict.pop("play_guid", None)
        pfx_dict.pop("pitcher_name", None)
        pfx_dict.pop("pitch_con", None)
        pfx_dict.pop("norm_ht", None)
        pfx_dict.pop("tstart", None)
        pfx_dict.pop("vystart", None)
        pfx_dict.pop("ftime", None)
        pfx_dict.pop("x0", None)
        pfx_dict.pop("y0", None)
        pfx_dict.pop("z0", None)
        pfx_dict.pop("vx0", None)
        pfx_dict.pop("vy0", None)
        pfx_dict.pop("vz0", None)
        pfx_dict.pop("ax", None)
        pfx_dict.pop("ay", None)
        pfx_dict.pop("az", None)
        pfx_dict.pop("tm_spin", None)
        pfx_dict.pop("sb", None)
        return pfx_dict


@accept_whitespaces
@dateformat(CSV_UTC)
@dataclass
class PitchFxCsvRow:
    id: int = None
    bb_game_id: str = None
    bbref_game_id: str = None
    pitch_app_id: str = None
    inning_id: str = None
    at_bat_id: str = None
    pitcher_id_mlb: int = None
    batter_id_mlb: int = None
    pitcher_id_bbref: str = None
    batter_id_bbref: str = None
    pitcher_team_id_bb: str = None
    opponent_team_id_bb: str = None
    p_throws: str = None
    stand: str = None
    pitch_id: int = None
    inning: int = None
    ab_total: int = None
    ab_count: int = None
    ab_id: int = None
    table_row_number: int = 0
    des: str = None
    strikes: int = 0
    balls: int = 0
    basic_type: str = None
    pdes: str = None
    mlbam_pitch_name: str = None
    start_speed: float = None
    spin: float = None
    zone_location: int = None
    sz_top: float = None
    sz_bot: float = None
    pfx_xdatafile: float = None
    pfx_zdatafile: float = None
    pfx_x: float = None
    pfx_z: float = None
    uncorrected_pfx_x: float = None
    uncorrected_pfx_z: float = None
    px: float = None
    pz: float = None
    pxold: float = None
    pzold: float = None
    park_sv_id: str = None
    game_start_time_utc: datetime = None
    time_pitch_thrown_utc: datetime = None
    seconds_since_game_start: int = 0
    has_zone_location: int = 0
    batter_did_swing: int = 0
    batter_made_contact: int = 0
    called_strike: int = 0
    swinging_strike: int = 0
    inside_strike_zone: int = 0
    outside_strike_zone: int = 0
    swing_inside_zone: int = 0
    swing_outside_zone: int = 0
    contact_inside_zone: int = 0
    contact_outside_zone: int = 0
    is_batted_ball: int = 0
    is_ground_ball: int = 0
    is_fly_ball: int = 0
    is_line_drive: int = 0
    is_pop_up: int = 0
    is_final_pitch_of_ab: int = 0
    ab_result_out: int = 0
    ab_result_hit: int = 0
    ab_result_single: int = 0
    ab_result_double: int = 0
    ab_result_triple: int = 0
    ab_result_homerun: int = 0
    ab_result_bb: int = 0
    ab_result_ibb: int = 0
    ab_result_k: int = 0
    ab_result_hbp: int = 0
    ab_result_error: int = 0
    ab_result_sac_hit: int = 0
    ab_result_sac_fly: int = 0
    ab_result_unclear: int = 0
    pitch_type_int: int = 0
    pbp_play_result: str = ""
    pbp_runs_outs_result: str = ""
    is_sp: int = 0
    is_rp: int = 0
    is_patched: int = 0
    is_invalid_ibb: int = 0
    is_out_of_sequence: int = 0



from dataclasses import dataclass


@dataclass
class BBRefStartingLineupSlot:
    """Batting order and defensive position for a single player in a starting lineup."""

    player_id_br: str = ""
    bat_order: str = ""
    def_position: str = ""

    def as_dict(self):
        """Convert lineup details to a dictionary."""
        return {
            "player_id_br": f"{self.player_id_br}",
            "bat_order": int(self.bat_order),
            "def_position": f"{self.def_position}",
        }



from datetime import datetime
from pathlib import Path
from typing import Dict, List, Tuple, Union

import vigorish.database as db
from vigorish.data.file_helper import FileHelper
from vigorish.data.html_storage import HtmlStorage
from vigorish.data.json_storage import JsonStorage
from vigorish.data.metrics import (
    BatStatsMetrics,
    PitchFxMetrics,
    PitchFxMetricsCollection,
    PitchStatsMetrics,
)
from vigorish.data.name_search import PlayerNameSearch
from vigorish.enums import DataSet, DefensePosition, PitchType, VigFile
from vigorish.util.regex import URL_ID_CONVERT_REGEX, URL_ID_REGEX
from vigorish.util.result import Result


class ScrapedData:
    def __init__(self, db_engine, db_session, config):
        self.db_engine = db_engine
        self.db_session = db_session
        self.config = config
        self.file_helper = FileHelper(config)
        self.html_storage = HtmlStorage(config, self.file_helper)
        self.json_storage = JsonStorage(config, self.file_helper)
        self.name_search = PlayerNameSearch(db_session)

    def get_local_folderpath(self, file_type, data_set, year):
        return self.file_helper.local_folderpath_dict[file_type][data_set].resolve(year=year)

    def get_s3_folderpath(self, file_type, data_set, year):  # pragma: no cover
        return self.file_helper.s3_folderpath_dict[file_type][data_set].resolve(year=year)

    def check_s3_bucket(self):  # pragma: no cover
        return self.file_helper.check_s3_bucket()

    def create_all_folderpaths(self, year):
        return self.file_helper.create_all_folderpaths(year)

    def save_html(self, data_set, url_id, html):
        return self.html_storage.save_html(data_set, url_id, html)

    def get_html(self, data_set, url_id):
        return self.html_storage.get_html(data_set, url_id)

    def save_json(self, data_set, parsed_data):
        return self.json_storage.save_json(data_set, parsed_data)

    def get_scraped_data(self, data_set, url_id, apply_patch_list=True):
        data = self.json_storage.get_scraped_data(data_set, url_id)
        if not apply_patch_list:
            return data
        result = self.apply_patch_list(data_set, url_id, data)
        return result.value if result.success else None

    def get_brooks_games_for_date(self, game_date, apply_patch_list=True):
        return self.get_scraped_data(DataSet.BROOKS_GAMES_FOR_DATE, game_date, apply_patch_list)

    def get_brooks_pitch_logs_for_game(self, bb_game_id):
        return self.get_scraped_data(DataSet.BROOKS_PITCH_LOGS, bb_game_id, apply_patch_list=False)

    def get_brooks_pitchfx_log(self, pitch_app_id, apply_patch_list=False):
        return self.get_scraped_data(DataSet.BROOKS_PITCHFX, pitch_app_id, apply_patch_list)

    def get_all_brooks_pitchfx_logs_for_game(self, bbref_game_id, apply_patch_list=True):
        pitchfx_logs = [
            self.get_brooks_pitchfx_log(pitch_app_id)
            for pitch_app_id in self.get_all_pitch_app_ids_with_pfx_data_for_game(bbref_game_id)
        ]
        return (
            Result.Fail(f"Failed to retrieve all pitchfx logs for game {bbref_game_id}")
            if not all(pfx_log for pfx_log in pitchfx_logs)
            else self.apply_patch_list(DataSet.BROOKS_PITCHFX, bbref_game_id, pitchfx_logs)
            if apply_patch_list
            else Result.Ok(pitchfx_logs)
        )

    def get_all_pitch_app_ids_with_pfx_data_for_game(self, bbref_game_id):
        return db.PitchAppScrapeStatus.get_all_pitch_app_ids_with_pfx_data_for_game(self.db_session, bbref_game_id)

    def get_bbref_games_for_date(self, game_date, apply_patch_list=True):
        return self.get_scraped_data(DataSet.BBREF_GAMES_FOR_DATE, game_date, apply_patch_list)

    def get_bbref_boxscore(self, bbref_game_id, apply_patch_list=True):
        return self.get_scraped_data(DataSet.BBREF_BOXSCORES, bbref_game_id, apply_patch_list)

    def save_patch_list(self, data_set, patch_list):
        return self.json_storage.save_patch_list(data_set, patch_list)

    def get_patch_list(self, data_set, url_id):
        return self.json_storage.get_patch_list(data_set, url_id)

    def apply_patch_list(self, data_set, url_id, data_to_patch):
        patch_list = self.get_patch_list(data_set, url_id)
        if not patch_list:
            return Result.Ok(data_to_patch)
        kwargs = self.get_data_required_for_patch_list(data_set, url_id)
        return patch_list.apply(data_to_patch, **kwargs)

    def get_data_required_for_patch_list(self, data_set, url_id):
        req_data_dict = {
            DataSet.BROOKS_GAMES_FOR_DATE: self.get_data_for_brooks_games_for_date_patch_list,
            DataSet.BROOKS_PITCHFX: self.get_data_for_brooks_pitchfx_logs_patch_list,
            DataSet.BBREF_GAMES_FOR_DATE: self.get_data_for_bbref_games_for_date_patch_list,
        }
        return req_data_dict[data_set](url_id) if data_set in req_data_dict else {}

    def get_data_for_brooks_games_for_date_patch_list(self, url_id):
        return {"db_session": self.db_session}

    def get_data_for_brooks_pitchfx_logs_patch_list(self, url_id):
        return {"boxscore": self.get_bbref_boxscore(url_id), "db_session": self.db_session}

    def get_data_for_bbref_games_for_date_patch_list(self, url_id):
        return {"db_session": self.db_session}

    def save_combined_game_data(self, combined_data):
        return self.json_storage.save_combined_game_data(combined_data)

    def get_combined_game_data(self, bbref_game_id):
        return self.json_storage.get_combined_game_data(bbref_game_id)

    def get_all_brooks_pitch_logs_for_date(self, game_date):
        brooks_game_ids = db.GameScrapeStatus.get_all_brooks_game_ids_for_date(self.db_session, game_date)
        pitch_logs = []
        for game_id in brooks_game_ids:
            pitch_log = self.get_brooks_pitch_logs_for_game(game_id)
            if not pitch_log:
                continue
            pitch_logs.append(pitch_log)
        return pitch_logs

    def get_scraped_ids_from_local_folder(self, file_type, data_set, year):
        folderpath = self.get_local_folderpath(file_type, data_set, year)
        url_ids = [file.stem for file in Path(folderpath).glob("*.*")]
        return self.validate_url_ids(file_type, data_set, url_ids)

    def validate_url_ids(self, file_type, data_set, url_ids):
        url_ids = filter(lambda x: URL_ID_REGEX[file_type][data_set].search(x), url_ids)
        url_ids = self.convert_url_ids(file_type, data_set, url_ids)
        return sorted(url_ids)

    def convert_url_ids(self, file_type, data_set, url_ids):
        if data_set not in URL_ID_CONVERT_REGEX[file_type]:
            return list(url_ids)
        convert_regex = URL_ID_CONVERT_REGEX[file_type][data_set]
        converted_url_ids = []
        for url_id in url_ids:
            match = convert_regex.search(url_id)
            if not match:
                raise ValueError(f"URL identifier is invalid: {url_id} ({data_set})")
            if file_type == VigFile.COMBINED_GAME_DATA:
                converted = match.groupdict()["bbref_game_id"]
            else:
                converted = self.convert_url_id_to_date(data_set, url_id, match.groupdict())
            converted_url_ids.append(converted)
        return converted_url_ids

    def convert_url_id_to_date(self, data_set, url_id, group_dict):
        try:
            year = int(group_dict["year"])
            month = int(group_dict["month"])
            day = int(group_dict["day"])
            return datetime(year, month, day)
        except Exception as e:
            error = f'Failed to parse date from url_id "{url_id} ({data_set})":\n{repr(e)}'
            raise ValueError(error)

    def get_scraped_ids_from_database(self, data_set, season):
        get_scraped_ids_from_db_dict = {
            DataSet.BROOKS_GAMES_FOR_DATE: self.get_all_brooks_scraped_dates_for_season,
            DataSet.BROOKS_PITCH_LOGS: self.get_all_scraped_brooks_game_ids_for_season,
            DataSet.BROOKS_PITCHFX: self.get_all_scraped_pitch_app_ids_for_season,
            DataSet.BBREF_GAMES_FOR_DATE: self.get_all_bbref_scraped_dates_for_season,
            DataSet.BBREF_BOXSCORES: self.get_all_scraped_bbref_game_ids_for_season,
        }
        return get_scraped_ids_from_db_dict[data_set](season)

    def get_all_brooks_scraped_dates_for_season(self, season):
        scraped_dates = db.DateScrapeStatus.get_all_brooks_scraped_dates_for_season(self.db_session, season.id)
        return sorted(scraped_dates)

    def get_all_scraped_brooks_game_ids_for_season(self, season):
        scraped_game_ids = db.GameScrapeStatus.get_all_scraped_brooks_game_ids_for_season(self.db_session, season.id)
        return sorted(scraped_game_ids)

    def get_all_scraped_pitch_app_ids_for_season(self, season):
        scraped_pitch_app_ids = db.PitchAppScrapeStatus.get_all_scraped_pitch_app_ids_for_season(
            self.db_session, season.id
        )
        return sorted(scraped_pitch_app_ids)

    def get_all_bbref_scraped_dates_for_season(self, season):
        scraped_dates = db.DateScrapeStatus.get_all_bbref_scraped_dates_for_season(self.db_session, season.id)
        return sorted(scraped_dates)

    def get_all_scraped_bbref_game_ids_for_season(self, season):
        scraped_game_ids = db.GameScrapeStatus.get_all_scraped_bbref_game_ids_for_season(self.db_session, season.id)
        return sorted(scraped_game_ids)

    def get_audit_report(self):
        all_seasons = db.Season.get_all_regular_seasons(self.db_session)
        total_games = db.Season.get_total_games_for_all_seasons(self.db_session)
        scraped = self.get_all_bbref_game_ids_eligible_for_audit(all_seasons)
        successful = self.get_all_bbref_game_ids_all_pitchfx_logs_are_valid(all_seasons)
        failed = self.get_all_bbref_game_ids_combined_data_fail(all_seasons)
        pfx_error = self.get_all_bbref_game_ids_pitchfx_error(all_seasons)
        invalid_pfx = self.get_all_bbref_game_ids_invalid_pitchfx(all_seasons)
        return {
            s.year: {
                "total_games": total_games[s.year],
                "scraped": scraped[s.year],
                "successful": successful[s.year],
                "failed": failed[s.year],
                "pfx_error": pfx_error[s.year],
                "invalid_pfx": invalid_pfx[s.year],
            }
            for s in all_seasons
            if scraped[s.year] or successful[s.year] or failed[s.year] or pfx_error[s.year] or invalid_pfx[s.year]
        }

    def get_all_bbref_game_ids_eligible_for_audit(self, all_seasons):
        return {
            s.year: db.Season_Game_PitchApp_View.get_all_bbref_game_ids_eligible_for_audit(self.db_engine, s.year)
            for s in all_seasons
        }

    def get_all_bbref_game_ids_combined_data_fail(self, all_seasons):
        return {s.year: s.get_all_bbref_game_ids_combined_data_fail() for s in all_seasons}

    def get_all_bbref_game_ids_pitchfx_error(self, all_seasons):
        return {
            s.year: db.Season_Game_PitchApp_View.get_all_bbref_game_ids_pitchfx_error(self.db_engine, s.year)
            for s in all_seasons
        }

    def get_all_bbref_game_ids_invalid_pitchfx(self, all_seasons):
        return {
            s.year: db.Season_Game_PitchApp_View.get_all_bbref_game_ids_invalid_pitchfx(self.db_engine, s.year)
            for s in all_seasons
        }

    def get_all_bbref_game_ids_all_pitchfx_logs_are_valid(self, all_seasons):
        return {
            s.year: db.Season_Game_PitchApp_View.get_all_bbref_game_ids_all_pitchfx_logs_are_valid(
                self.db_engine, s.year
            )
            for s in all_seasons
        }

    def get_all_seasons_with_data_for_player(self, player_mlb_id: int) -> List[db.Season]:
        seasons_with_pitch_data = db.PitchStats.get_all_seasons_with_data_for_player(self.db_session, player_mlb_id)
        seasons_with_bat_data = db.BatStats.get_all_seasons_with_data_for_player(self.db_session, player_mlb_id)
        seasons_with_player_data = list(set(list(seasons_with_pitch_data) + list(seasons_with_bat_data)))
        return sorted(seasons_with_player_data, key=lambda x: x.year) if seasons_with_player_data else []

    def player_name_search(self, query):
        return self.name_search.fuzzy_match(query)

    # TEAM PITCH STATS

    def get_pitch_stats_for_team(self, team_id_bbref: str, year: int) -> PitchStatsMetrics:
        return db.Team_PitchStats_By_Year_View.get_pitch_stats_for_team(self.db_engine, team_id_bbref, year)

    def get_pitch_stats_for_sp_for_team(self, team_id_bbref: str, year: int) -> PitchStatsMetrics:
        return db.Team_PitchStats_SP_By_Year_View.get_pitch_stats_for_sp_for_team(self.db_engine, team_id_bbref, year)

    def get_pitch_stats_for_rp_for_team(self, team_id_bbref: str, year: int) -> PitchStatsMetrics:
        return db.Team_PitchStats_RP_By_Year_View.get_pitch_stats_for_rp_for_team(self.db_engine, team_id_bbref, year)

    def get_pitch_stats_by_year_for_team(self, team_id_bbref: str) -> Dict[int, PitchStatsMetrics]:
        team_stats = db.Team_PitchStats_By_Year_View.get_pitch_stats_by_year_for_team(self.db_engine, team_id_bbref)
        return {s.year: s for s in team_stats}

    def get_pitch_stats_for_sp_by_year_for_team(self, team_id_bbref: str) -> Dict[int, PitchStatsMetrics]:
        team_stats = db.Team_PitchStats_SP_By_Year_View.get_pitch_stats_for_sp_by_year_for_team(
            self.db_engine, team_id_bbref
        )
        return {s.year: s for s in team_stats}

    def get_pitch_stats_for_rp_by_year_for_team(self, team_id_bbref: str) -> Dict[int, PitchStatsMetrics]:
        team_stats = db.Team_PitchStats_RP_By_Year_View.get_pitch_stats_for_rp_by_year_for_team(
            self.db_engine, team_id_bbref
        )
        return {s.year: s for s in team_stats}

    def get_pitch_stats_by_player_for_team(self, team_id_bbref: str, year: int) -> List[PitchStatsMetrics]:
        pitch_stats = db.Team_PitchStats_By_Player_By_Year_View.get_pitch_stats_by_player_for_team(
            self.db_engine, team_id_bbref, year
        )
        return sorted(pitch_stats, key=lambda x: x.re24_pitch, reverse=True)

    def get_pitch_stats_for_sp_by_player_for_team(self, team_id_bbref: str, year: int) -> List[PitchStatsMetrics]:
        pitch_stats = db.Team_PitchStats_SP_By_Player_By_Year_View.get_pitch_stats_for_sp_by_player_for_team(
            self.db_engine, team_id_bbref, year
        )
        return sorted(pitch_stats, key=lambda x: x.re24_pitch, reverse=True)

    def get_pitch_stats_for_rp_by_player_for_team(self, team_id_bbref: str, year: int) -> List[PitchStatsMetrics]:
        pitch_stats = db.Team_PitchStats_RP_By_Player_By_Year_View.get_pitch_stats_for_rp_by_player_for_team(
            self.db_engine, team_id_bbref, year
        )
        return sorted(pitch_stats, key=lambda x: x.re24_pitch, reverse=True)

    def get_pitch_stats_for_season_for_all_teams(self, year: int) -> Dict[str, PitchStatsMetrics]:
        team_stats = db.Team_PitchStats_By_Year_View.get_pitch_stats_for_season_for_all_teams(self.db_engine, year)
        return _sort_and_map_team_pitch_stats(team_stats)

    def get_pitch_stats_for_sp_for_season_for_all_teams(self, year: int) -> Dict[str, PitchStatsMetrics]:
        taem_stats = db.Team_PitchStats_SP_By_Year_View.get_pitch_stats_for_sp_for_season_for_all_teams(
            self.db_engine, year
        )
        return _sort_and_map_team_pitch_stats(taem_stats)

    def get_pitch_stats_for_rp_for_season_for_all_teams(self, year: int) -> Dict[str, PitchStatsMetrics]:
        team_stats = db.Team_PitchStats_RP_By_Year_View.get_pitch_stats_for_rp_for_season_for_all_teams(
            self.db_engine, year
        )
        return _sort_and_map_team_pitch_stats(team_stats)

    # TEAM BAT STATS

    def get_bat_stats_for_team(self, team_id_bbref: str, year: int) -> BatStatsMetrics:
        return db.Team_BatStats_By_Year_View.get_bat_stats_for_team(self.db_engine, team_id_bbref, year)

    def get_bat_stats_by_lineup_spot_for_team(self, team_id_bbref: str, year: int) -> List[BatStatsMetrics]:
        return db.Team_BatStats_By_BatOrder_By_Year.get_bat_stats_by_lineup_spot_for_team(
            self.db_engine, team_id_bbref, year
        )

    def get_bat_stats_by_defpos_for_team(self, team_id_bbref: str, year: int) -> List[BatStatsMetrics]:
        return db.Team_BatStats_By_DefPosition_By_Year.get_bat_stats_by_defpos_for_team(
            self.db_engine, team_id_bbref, year
        )

    def get_bat_stats_for_starters_for_team(self, team_id_bbref: str, year: int) -> BatStatsMetrics:
        return db.Team_BatStats_For_Starters_By_Year.get_bat_stats_for_starters_for_team(
            self.db_engine, team_id_bbref, year
        )

    def get_bat_stats_for_subs_for_team(self, team_id_bbref: str, year: int) -> BatStatsMetrics:
        return db.Team_BatStats_For_Subs_By_Year.get_bat_stats_for_subs_for_team(self.db_engine, team_id_bbref, year)

    def get_bat_stats_by_year_for_team(self, team_id_bbref: str) -> Dict[int, BatStatsMetrics]:
        team_stats = db.Team_BatStats_By_Year_View.get_bat_stats_by_year_for_team(self.db_engine, team_id_bbref)
        return {s.year: s for s in team_stats}

    def get_bat_stats_for_lineup_spot_by_year_for_team(
        self, lineup_spot: int, team_id_bbref: str
    ) -> Dict[int, BatStatsMetrics]:
        team_stats = db.Team_BatStats_By_BatOrder_By_Year.get_bat_stats_for_lineup_spot_by_year_for_team(
            self.db_engine, lineup_spot, team_id_bbref
        )
        return {s.year: s for s in team_stats}

    def get_bat_stats_for_defpos_by_year_for_team(
        self, def_position: DefensePosition, team_id_bbref: str
    ) -> Dict[int, BatStatsMetrics]:
        team_stats = db.Team_BatStats_By_DefPosition_By_Year.get_bat_stats_for_defpos_by_year_for_team(
            self.db_engine, int(def_position), team_id_bbref
        )
        return {s.year: s for s in team_stats}

    def get_bat_stats_for_starters_by_year_for_team(self, team_id_bbref: str) -> Dict[int, BatStatsMetrics]:
        team_stats = db.Team_BatStats_For_Starters_By_Year.get_bat_stats_for_starters_by_year_for_team(
            self.db_engine, team_id_bbref
        )
        return {s.year: s for s in team_stats}

    def get_bat_stats_for_subs_by_year_for_team(self, team_id_bbref: str) -> Dict[int, BatStatsMetrics]:
        team_stats = db.Team_BatStats_For_Subs_By_Year.get_bat_stats_for_subs_by_year_for_team(
            self.db_engine, team_id_bbref
        )
        return {s.year: s for s in team_stats}

    def get_bat_stats_by_player_for_team(self, team_id_bbref: str, year: int) -> List[BatStatsMetrics]:
        team_stats = db.Team_BatStats_By_Player_By_Year_View.get_bat_stats_by_player_for_team(
            self.db_engine, team_id_bbref, year
        )
        return sorted(team_stats, key=lambda x: x.plate_appearances, reverse=True)

    def get_bat_stats_for_lineup_spot_by_player_for_team(
        self, lineup_spot: int, team_id_bbref: str, year: int
    ) -> List[BatStatsMetrics]:
        team_stats = db.Team_BatStats_By_BatOrder_By_Player_By_Year.get_bat_stats_for_lineup_spot_by_player_for_team(
            self.db_engine, lineup_spot, team_id_bbref, year
        )
        return sorted(team_stats, key=lambda x: x.plate_appearances, reverse=True)

    def get_bat_stats_for_defpos_by_player_for_team(
        self, def_position: DefensePosition, team_id_bbref: str, year: int
    ) -> List[BatStatsMetrics]:
        team_stats = db.Team_BatStats_By_DefPosition_By_Player_By_Year.get_bat_stats_for_defpos_by_player_for_team(
            self.db_engine, int(def_position), team_id_bbref, year
        )
        return sorted(team_stats, key=lambda x: x.plate_appearances, reverse=True)

    def get_bat_stats_for_starters_by_player_for_team(self, team_id_bbref: str, year: int) -> List[BatStatsMetrics]:
        team_stats = db.Team_BatStats_For_Starters_By_Player_By_Year.get_bat_stats_for_starters_by_player_for_team(
            self.db_engine, team_id_bbref, year
        )
        return sorted(team_stats, key=lambda x: x.plate_appearances, reverse=True)

    def get_bat_stats_for_subs_by_player_for_team(self, team_id_bbref: str, year: int) -> List[BatStatsMetrics]:
        team_stats = db.Team_BatStats_For_Subs_By_Player_By_Year.get_bat_stats_for_subs_by_player_for_team(
            self.db_engine, team_id_bbref, year
        )
        return sorted(team_stats, key=lambda x: x.plate_appearances, reverse=True)

    def get_bat_stats_for_season_for_all_teams(self, year: int) -> Dict[str, BatStatsMetrics]:
        team_stats = db.Team_BatStats_By_Year_View.get_bat_stats_for_season_for_all_teams(self.db_engine, year)
        return _sort_and_map_team_bat_stats(team_stats)

    def get_bat_stats_for_lineup_spot_for_season_for_all_teams(
        self, lineup_spot: int, year: int
    ) -> Dict[str, BatStatsMetrics]:
        team_batstats = db.Team_BatStats_By_BatOrder_By_Year.get_bat_stats_for_lineup_spot_for_season_for_all_teams(
            self.db_engine, lineup_spot, year
        )
        return _sort_and_map_team_bat_stats(team_batstats)

    def get_bat_stats_for_defpos_for_season_for_all_teams(
        self, def_position: DefensePosition, year: int
    ) -> Dict[str, BatStatsMetrics]:
        team_batstats = db.Team_BatStats_By_DefPosition_By_Year.get_bat_stats_for_defpos_for_season_for_all_teams(
            self.db_engine, int(def_position), year
        )
        return _sort_and_map_team_bat_stats(team_batstats)

    def get_bat_stats_for_starters_for_season_for_all_teams(self, year: int) -> Dict[str, BatStatsMetrics]:
        team_batstats = db.Team_BatStats_For_Starters_By_Year.get_bat_stats_for_starters_for_season_for_all_teams(
            self.db_engine, year
        )
        return _sort_and_map_team_bat_stats(team_batstats)

    def get_bat_stats_for_subs_for_season_for_all_teams(self, year: int) -> Dict[str, BatStatsMetrics]:
        team_batstats = db.Team_BatStats_For_Subs_By_Year.get_bat_stats_for_subs_for_season_for_all_teams(
            self.db_engine, year
        )
        return _sort_and_map_team_bat_stats(team_batstats)

    # PLAYER PITCH STATS

    def get_pitch_stats_for_career_for_player(self, mlb_id: int) -> PitchStatsMetrics:
        return db.PitchStats_All_View.get_pitch_stats_for_career_for_player(self.db_engine, mlb_id)

    def get_pitch_stats_as_sp_for_player(self, mlb_id: int) -> PitchStatsMetrics:
        return db.PitchStats_SP_View.get_pitch_stats_as_sp_for_player(self.db_engine, mlb_id)

    def get_pitch_stats_as_rp_for_player(self, mlb_id: int) -> PitchStatsMetrics:
        return db.PitchStats_RP_View.get_pitch_stats_as_rp_for_player(self.db_engine, mlb_id)

    def get_pitch_stats_by_year_for_player(self, mlb_id: int) -> List[PitchStatsMetrics]:
        return db.PitchStats_By_Year_View.get_pitch_stats_by_year_for_player(self.db_engine, mlb_id)

    def get_pitch_stats_by_team_for_player(self, mlb_id: int) -> List[PitchStatsMetrics]:
        return db.PitchStats_By_Team_View.get_pitch_stats_by_team_for_player(self.db_engine, mlb_id)

    def get_pitch_stats_by_team_by_year_for_player(self, mlb_id: int) -> List[PitchStatsMetrics]:
        return db.PitchStats_By_Team_By_Year_View.get_pitch_stats_by_team_by_year_for_player(self.db_engine, mlb_id)

    def get_pitch_stats_by_opp_team_for_player(self, mlb_id: int) -> List[PitchStatsMetrics]:
        return db.PitchStats_By_Opp_Team_View.get_pitch_stats_by_opp_team_for_player(self.db_engine, mlb_id)

    def get_pitch_stats_by_opp_team_by_year_for_player(self, mlb_id: int) -> List[PitchStatsMetrics]:
        return db.PitchStats_By_Opp_Team_By_Year_View.get_pitch_stats_by_opp_team_by_year_for_player(
            self.db_engine, mlb_id
        )

    # PITCHER PITCHFX STATS

    def get_career_pfx_metrics_for_pitcher(self, mlb_id: int) -> PitchFxMetricsCollection:
        return PitchFxMetricsCollection.from_query_results(
            db.Pitcher_PitchFx_All_View.get_pfx_metrics_for_career_for_pitcher(self.db_engine, mlb_id),
            db.Pitcher_PitchType_All_View.get_pfx_metrics_for_career_for_pitcher(self.db_engine, mlb_id),
        )

    def calculate_pitch_type_percentiles(
        self, p_throws: str, pfx_metrics: PitchFxMetrics
    ) -> Dict[str, Union[PitchType, Tuple[float, float]]]:
        return db.PitchTypePercentile.calculate_pitch_type_percentiles(self.db_session, p_throws, pfx_metrics)

    def get_career_pfx_metrics_vs_rhb_for_pitcher(self, mlb_id: int) -> PitchFxMetricsCollection:
        return PitchFxMetricsCollection.from_query_results(
            db.Pitcher_PitchFx_vs_RHB_View.get_pfx_metrics_vs_rhb_for_pitcher(self.db_engine, mlb_id),
            db.Pitcher_PitchType_vs_RHB_View.get_pfx_metrics_vs_rhb_for_pitcher(self.db_engine, mlb_id),
        )

    def get_career_pfx_metrics_vs_lhb_for_pitcher(self, mlb_id: int) -> PitchFxMetricsCollection:
        return PitchFxMetricsCollection.from_query_results(
            db.Pitcher_PitchFx_vs_LHB_View.get_pfx_metrics_vs_lhb_for_pitcher(self.db_engine, mlb_id),
            db.Pitcher_PitchType_vs_LHB_View.get_pfx_metrics_vs_lhb_for_pitcher(self.db_engine, mlb_id),
        )

    def get_pfx_metrics_for_year_for_pitcher(self, mlb_id: int, year: int) -> PitchFxMetricsCollection:
        return PitchFxMetricsCollection.from_query_results(
            db.Pitcher_PitchFx_All_By_Year_View.get_pfx_metrics_by_year_for_pitcher(self.db_engine, mlb_id, year),
            db.Pitcher_PitchType_All_By_Year_View.get_pfx_metrics_by_year_for_pitcher(self.db_engine, mlb_id, year),
        )

    def get_pfx_metrics_for_year_vs_rhb_for_pitcher(self, mlb_id: int, year: int) -> PitchFxMetricsCollection:
        return PitchFxMetricsCollection.from_query_results(
            db.Pitcher_PitchFx_vs_RHB_By_Year_View.get_pfx_metrics_by_year_for_pitcher(self.db_engine, mlb_id, year),
            db.Pitcher_PitchType_vs_RHB_By_Year_View.get_pfx_metrics_by_year_for_pitcher(self.db_engine, mlb_id, year),
        )

    def get_pfx_metrics_for_year_vs_lhb_for_pitcher(self, mlb_id: int, year: int) -> PitchFxMetricsCollection:
        return PitchFxMetricsCollection.from_query_results(
            db.Pitcher_PitchFx_vs_LHB_By_Year_View.get_pfx_metrics_by_year_for_pitcher(self.db_engine, mlb_id, year),
            db.Pitcher_PitchType_vs_LHB_By_Year_View.get_pfx_metrics_by_year_for_pitcher(self.db_engine, mlb_id, year),
        )

    def get_pfx_metrics_for_game_for_pitcher(self, mlb_id: int, bbref_game_id: str) -> PitchFxMetricsCollection:
        return PitchFxMetricsCollection.from_query_results(
            db.Pitcher_PitchFx_For_Game_All_View.get_pfx_metrics_for_game_for_pitcher(
                self.db_engine, mlb_id, bbref_game_id
            ),
            db.Pitcher_PitchType_For_Game_All_View.get_pfx_metrics_for_game_for_pitcher(
                self.db_engine, mlb_id, bbref_game_id
            ),
        )

    def get_pfx_metrics_for_game_vs_rhb_for_pitcher(self, mlb_id: int, bbref_game_id: str) -> PitchFxMetricsCollection:
        return PitchFxMetricsCollection.from_query_results(
            db.Pitcher_PitchFx_For_Game_vs_RHB_View.get_pfx_metrics_for_game_vs_rhb_for_pitcher(
                self.db_engine, mlb_id, bbref_game_id
            ),
            db.Pitcher_PitchType_For_Game_vs_RHB_View.get_pfx_metrics_for_game_vs_rhb_for_pitcher(
                self.db_engine, mlb_id, bbref_game_id
            ),
        )

    def get_pfx_metrics_for_game_vs_lhb_for_pitcher(self, mlb_id: int, bbref_game_id: str) -> PitchFxMetricsCollection:
        return PitchFxMetricsCollection.from_query_results(
            db.Pitcher_PitchFx_For_Game_vs_LHB_View.get_pfx_metrics_for_game_vs_lhb_for_pitcher(
                self.db_engine, mlb_id, bbref_game_id
            ),
            db.Pitcher_PitchType_For_Game_vs_LHB_View.get_pfx_metrics_for_game_vs_lhb_for_pitcher(
                self.db_engine, mlb_id, bbref_game_id
            ),
        )

    # PLAYER BAT STATS

    def get_bat_stats_for_career_for_player(self, mlb_id: int) -> BatStatsMetrics:
        return db.BatStats_All_View.get_bat_stats_for_career_for_player(self.db_engine, mlb_id)

    def get_bat_stats_by_year_for_player(self, mlb_id: int) -> List[BatStatsMetrics]:
        return db.BatStats_By_Year_View.get_bat_stats_by_year_for_player(self.db_engine, mlb_id)

    def get_bat_stats_by_team_for_player(self, mlb_id: int) -> List[BatStatsMetrics]:
        return db.BatStats_By_Team_View.get_bat_stats_by_team_for_player(self.db_engine, mlb_id)

    def get_bat_stats_by_team_by_year_for_player(self, mlb_id: int) -> List[BatStatsMetrics]:
        return db.BatStats_By_Team_Year_View.get_bat_stats_by_team_by_year_for_player(self.db_engine, mlb_id)

    def get_bat_stats_by_opp_for_player(self, mlb_id: int) -> List[BatStatsMetrics]:
        return db.BatStats_By_Opp_Team_View.get_bat_stats_by_opp_for_player(self.db_engine, mlb_id)

    def get_bat_stats_by_opp_by_year_for_player(self, mlb_id: int) -> List[BatStatsMetrics]:
        return db.BatStats_By_Opp_Team_Year_View.get_bat_stats_by_opp_by_year_for_player(self.db_engine, mlb_id)

    # BATTER PITCHFX STATS

    def get_pfx_metrics_for_career_for_batter(self, mlb_id: int) -> PitchFxMetricsCollection:
        return PitchFxMetricsCollection.from_query_results(
            db.Batter_PitchFx_All_View.get_pfx_metrics_for_career_for_batter(self.db_engine, mlb_id),
            db.Batter_PitchType_All_View.get_pfx_metrics_for_career_for_batter(self.db_engine, mlb_id),
        )

    def get_pfx_metrics_vs_rhp_as_rhb_for_career_for_batter(self, mlb_id: int) -> PitchFxMetricsCollection:
        return PitchFxMetricsCollection.from_query_results(
            db.Batter_PitchFx_vs_RHP_as_RHB_Career_View.get_pfx_metrics_vs_rhp_as_rhb_for_career_for_batter(
                self.db_engine, mlb_id
            ),
            db.Batter_PitchType_vs_RHP_as_RHB_Career_View.get_pfx_metrics_vs_rhp_as_rhb_for_career_for_batter(
                self.db_engine, mlb_id
            ),
        )

    def get_pfx_metrics_vs_rhp_as_lhb_for_career_for_batter(self, mlb_id: int) -> PitchFxMetricsCollection:
        return PitchFxMetricsCollection.from_query_results(
            db.Batter_PitchFx_vs_RHP_as_LHB_Career_View.get_pfx_metrics_vs_rhp_as_lhb_for_career_for_batter(
                self.db_engine, mlb_id
            ),
            db.Batter_PitchType_vs_RHP_as_LHB_Career_View.get_pfx_metrics_vs_rhp_as_lhb_for_career_for_batter(
                self.db_engine, mlb_id
            ),
        )

    def get_pfx_metrics_vs_lhp_as_lhb_for_career_for_batter(self, mlb_id: int) -> PitchFxMetricsCollection:
        return PitchFxMetricsCollection.from_query_results(
            db.Batter_PitchFx_vs_LHP_as_LHB_Career_View.get_pfx_metrics_vs_lhp_as_lhb_for_career_for_batter(
                self.db_engine, mlb_id
            ),
            db.Batter_PitchType_vs_LHP_as_LHB_Career_View.get_pfx_metrics_vs_lhp_as_lhb_for_career_for_batter(
                self.db_engine, mlb_id
            ),
        )

    def get_pfx_metrics_vs_lhp_as_rhb_for_career_for_batter(self, mlb_id: int) -> PitchFxMetricsCollection:
        return PitchFxMetricsCollection.from_query_results(
            db.Batter_PitchFx_vs_LHP_as_RHB_Career_View.get_pfx_metrics_vs_lhp_as_rhb_for_career_for_batter(
                self.db_engine, mlb_id
            ),
            db.Batter_PitchType_vs_LHP_as_RHB_Career_View.get_pfx_metrics_vs_lhp_as_rhb_for_career_for_batter(
                self.db_engine, mlb_id
            ),
        )

    def get_pfx_metrics_for_year_for_batter(self, mlb_id: int, year: int) -> PitchFxMetricsCollection:
        return PitchFxMetricsCollection.from_query_results(
            db.Batter_PitchFx_By_Year_View.get_pfx_metrics_by_year_for_batter(self.db_engine, mlb_id, year),
            db.Batter_PitchType_By_Year_View.get_pfx_metrics_by_year_for_batter(self.db_engine, mlb_id, year),
        )

    def get_pfx_metrics_for_game_for_batter(self, mlb_id: int, bbref_game_id: str) -> PitchFxMetricsCollection:
        return PitchFxMetricsCollection.from_query_results(
            db.Batter_PitchFx_For_Game_All_View.get_pfx_metrics_for_game_for_batter(
                self.db_engine, mlb_id, bbref_game_id
            ),
            db.Batter_PitchType_For_Game_All_View.get_pfx_metrics_for_game_for_batter(
                self.db_engine, mlb_id, bbref_game_id
            ),
        )

    def get_pfx_metrics_vs_rhp_as_rhb_for_game_for_batter(
        self, mlb_id: int, bbref_game_id: str
    ) -> PitchFxMetricsCollection:
        return PitchFxMetricsCollection.from_query_results(
            db.Batter_PitchFx_For_Game_vs_RHP_as_RHB_View.get_pfx_metrics_vs_rhp_as_rhb_for_game_for_batter(
                self.db_engine, mlb_id, bbref_game_id
            ),
            db.Batter_PitchType_For_Game_vs_RHP_as_RHB_View.get_pfx_metrics_vs_rhp_as_rhb_for_game_for_batter(
                self.db_engine, mlb_id, bbref_game_id
            ),
        )

    def get_pfx_metrics_vs_rhp_as_lhb_for_game_for_batter(
        self, mlb_id: int, bbref_game_id: str
    ) -> PitchFxMetricsCollection:
        return PitchFxMetricsCollection.from_query_results(
            db.Batter_PitchFx_For_Game_vs_RHP_as_LHB_View.get_pfx_metrics_vs_rhp_as_lhb_for_game_for_batter(
                self.db_engine, mlb_id, bbref_game_id
            ),
            db.Batter_PitchType_For_Game_vs_RHP_as_LHB_View.get_pfx_metrics_vs_rhp_as_lhb_for_game_for_batter(
                self.db_engine, mlb_id, bbref_game_id
            ),
        )

    def get_pfx_metrics_vs_lhp_as_lhb_for_game_for_batter(
        self, mlb_id: int, bbref_game_id: str
    ) -> PitchFxMetricsCollection:
        return PitchFxMetricsCollection.from_query_results(
            db.Batter_PitchFx_For_Game_vs_LHP_as_LHB_View.get_pfx_metrics_vs_lhp_as_lhb_for_game_for_batter(
                self.db_engine, mlb_id, bbref_game_id
            ),
            db.Batter_PitchType_For_Game_vs_LHP_as_LHB_View.get_pfx_metrics_vs_lhp_as_lhb_for_game_for_batter(
                self.db_engine, mlb_id, bbref_game_id
            ),
        )

    def get_pfx_metrics_vs_lhp_as_rhb_for_game_for_batter(
        self, mlb_id: int, bbref_game_id: str
    ) -> PitchFxMetricsCollection:
        return PitchFxMetricsCollection.from_query_results(
            db.Batter_PitchFx_For_Game_vs_LHP_as_RHB_View.get_pfx_metrics_vs_lhp_as_rhb_for_game_for_batter(
                self.db_engine, mlb_id, bbref_game_id
            ),
            db.Batter_PitchType_For_Game_vs_LHP_as_RHB_View.get_pfx_metrics_vs_lhp_as_rhb_for_game_for_batter(
                self.db_engine, mlb_id, bbref_game_id
            ),
        )


def _sort_and_map_stats_list(stats_list, sort_attr, group_attr):
    stats_list.sort(key=lambda x: getattr(x, sort_attr), reverse=True)
    return {getattr(s, group_attr): s for s in map(_set_team_rank, enumerate(stats_list, start=1))}


def _sort_and_map_team_bat_stats(team_stats_list):
    return _sort_and_map_stats_list(team_stats_list, sort_attr="ops", group_attr="team_id_bbref")


def _sort_and_map_team_pitch_stats(team_stats_list):
    return _sort_and_map_stats_list(team_stats_list, sort_attr="re24_pitch", group_attr="team_id_bbref")


def _set_team_rank(rank_pair):
    (rank, stats) = rank_pair
    stats.rank = rank
    return stats



"""Update bbref_player_id_map.json file."""
import subprocess

from getch import pause
from halo import Halo

from vigorish.cli.components.util import (
    get_random_cli_color,
    get_random_dots_spinner,
    print_heading,
)
from vigorish.cli.components.viewers import DictListTableViewer
from vigorish.cli.menu_item import MenuItem
from vigorish.constants import EMOJIS
from vigorish.tasks import UpdatePlayerIdMapTask
from vigorish.util.result import Result


class UpdatePlayerIdMap(MenuItem):
    def __init__(self, app):
        super().__init__(app)
        self.id_map_task = UpdatePlayerIdMapTask(self.app)
        self.menu_item_text = " Update Player ID Map"
        self.menu_item_emoji = EMOJIS.get("TABBED_FILES")

    def launch(self):
        subprocess.run(["clear"])
        print_heading("Update Player ID Map", fg="bright_yellow")
        spinner = Halo(spinner=get_random_dots_spinner(), color=get_random_cli_color())
        spinner.text = "Checking BBRef player data..."
        spinner.start()
        result = self.id_map_task.execute()
        if result.failure:
            spinner.stop()
            return result
        spinner.succeed("Player ID map was successfully updated!")
        new_player_ids = result.value or []
        if not new_player_ids:
            pause(message="Press any key to continue...")
            return Result.Ok(new_player_ids)
        heading = "Update Player ID Map: Results"
        message = f"{len(new_player_ids)} new players were added to the ID map\n"
        table_viewer = DictListTableViewer(
            dict_list=new_player_ids,
            prompt="Press Enter to return to previous menu",
            confirm_only=True,
            table_color="bright_yellow",
            heading=heading,
            heading_color="bright_yellow",
            message=message,
            message_color="blue",
        )
        table_viewer.launch()
        return Result.Ok(new_player_ids)



"""Menu item that allows the user to create a record of a new scrape job."""
import subprocess

from vigorish.cli.components import (
    data_sets_prompt,
    DateInput,
    JobNameInput,
    print_heading,
    print_message,
    validate_scrape_dates,
    yes_no_prompt,
)
from vigorish.cli.menu_item import MenuItem
from vigorish.constants import DATA_SET_TO_NAME_MAP, EMOJIS
from vigorish.scrape.job_runner import JobRunner
from vigorish.util.dt_format_strings import DATE_ONLY_2
from vigorish.util.result import Result


class CreateJob(MenuItem):
    def __init__(self, app):
        # TODO: Improve Create Job UX:
        #  - Create new input type that is initialized with a list of valid values
        #  - Order list of values and the selected value can be changed with <- and -> arrow keys
        #  - This input type does not replace the options prompt with numbered items
        #  - This should be used for numbers and dates, like a slider element
        #  - In this menu, it will be used to select start/end dates
        #    - Separate inputs for month and day (user will select year first before data sets)
        super().__init__(app)
        self.menu_item_text = "Create New Job"
        self.menu_item_emoji = EMOJIS.get("KNIFE", "")
        self.menu_heading = self._menu_item_text

    def launch(self):
        job_confirmed = False
        data_sets = []
        start = None
        end = None
        job_name = None
        prompt = "Select all data sets to scrape:"
        while not job_confirmed:
            heading = self.get_menu_heading("Select Data Sets")
            data_sets = data_sets_prompt(heading, prompt, checked_data_sets=data_sets)
            dates_validated = False
            while not dates_validated:
                start = self.prompt_user_scrape_start_date(start)
                end = self.prompt_user_scrape_stop_date(end)
                result = validate_scrape_dates(self.db_session, start, end)
                if result.failure:
                    continue
                dates_validated = True
            job_name = self.prompt_user_job_name(job_name)
            job_confirmed = self.confirm_job_details(data_sets, start, end, job_name)

        new_job = self.app.create_scrape_job(data_sets, start, end, job_name).value
        subprocess.run(["clear"])
        if yes_no_prompt(prompt="\nWould you like to begin executing this job?"):
            job_runner = JobRunner(app=self.app, db_job=new_job)
            result = job_runner.execute()
            if result.failure:
                return result
        return Result.Ok(self.exit_menu)

    def prompt_user_scrape_start_date(self, start_date):
        date_prompt = DateInput(
            prompt="start date: ",
            heading="Enter date to START scraping",
            heading_color="bright_yellow",
            default=start_date,
        )
        return date_prompt.launch()

    def prompt_user_scrape_stop_date(self, stop_date):
        date_prompt = DateInput(
            prompt="stop date: ",
            heading="Enter date to STOP scraping",
            heading_color="bright_yellow",
            default=stop_date,
        )
        return date_prompt.launch()

    def prompt_user_job_name(self, job_name):
        name_prompt = JobNameInput(
            prompt="job name: ",
            heading="Enter a name for this job (Optional)",
            heading_color="bright_yellow",
            message="This value is optional, press ENTER to use the default value",
            default=job_name,
        )
        return name_prompt.launch()

    def confirm_job_details(self, data_sets, start_date, end_date, job_name):
        subprocess.run(["clear"])
        heading = "Confirm job details"
        data_set_space = "\n\t      "
        confirm_job_name = ""
        if job_name:
            confirm_job_name = f"Job Name....: {job_name}\n"
        job_details = (
            f"{confirm_job_name}"
            f"Start date..: {start_date.strftime(DATE_ONLY_2)}\n"
            f"End Date....: {end_date.strftime(DATE_ONLY_2)}\n"
            f"Data Sets...: {data_set_space.join([DATA_SET_TO_NAME_MAP[ds] for ds in data_sets])}"
        )
        print_heading(heading, fg="bright_yellow")
        print_message(job_details, wrap=False, fg="bright_yellow")
        return yes_no_prompt(prompt="\nAre the details above correct?")



"""Utility functions that transform and/or produce string values."""
import re
from datetime import datetime

from rapidfuzz import process

from vigorish.constants import BB_BR_TEAM_ID_MAP, BR_BB_TEAM_ID_MAP
from vigorish.util.dt_format_strings import DATE_ONLY_TABLE_ID
from vigorish.util.list_helpers import flatten_list2d
from vigorish.util.regex import (
    AT_BAT_ID_REGEX,
    BB_GAME_ID_REGEX,
    BBREF_BOXSCORE_URL_REGEX,
    BBREF_GAME_ID_REGEX,
    INNING_LABEL_REGEX,
    PITCH_APP_REGEX,
    TIMESTAMP_REGEX,
)
from vigorish.util.result import Result

ELLIPSIS = b"\xe2\x80\xa6".decode("utf-8")
NEWLINE_REGEX = re.compile(r"\n")
WORD_REGEX = re.compile(r"(?:\s|\b)?(?P<word>[\w\S]+)(?:\s|\B)?")


def fuzzy_match(query, mapped_choices, limit=10, score_cutoff=88):
    best_matches = [
        {"match": match, "score": score, "result": result}
        for (match, score, result) in process.extract(query, mapped_choices, limit=limit, score_cutoff=score_cutoff)
    ]
    return best_matches or [
        {"match": match, "score": score, "result": result}
        for (match, score, result) in process.extract(query, mapped_choices, limit=limit)
    ]


def ellipsize(input_str, max_len):
    if len(input_str) <= max_len:
        return input_str
    trunc = f"{input_str[:max_len - 1]} {ELLIPSIS}"
    for match in reversed(list(WORD_REGEX.finditer(input_str))):
        if match.end("word") > max_len:
            continue
        trunc = f"{input_str[:match.end('word')]} {ELLIPSIS}"
        break
    return trunc


def wrap_text(input_str, max_len):
    input_strings = [s for s in input_str.split("\n") if s]
    wrapped = flatten_list2d([_wrap_string(s, max_len) for s in input_strings])
    wrapped = _replace_newlines(input_str, wrapped)
    return "\n".join(wrapped)


def _wrap_string(s, max_len):
    substrings = []
    while True:
        if len(s) <= max_len:
            substrings.append(s)
            break
        (wrapped, s) = _word_wrap(s, max_len)
        substrings.append(wrapped)
    return substrings


def _word_wrap(s, max_len):
    last_word_boundary = max_len
    for match in WORD_REGEX.finditer(s):
        if match.end("word") > max_len:
            break
        last_word_boundary = match.end("word") + 1
    wrapped = s[:last_word_boundary]
    s = s[last_word_boundary:]
    return (wrapped.strip(), s.strip())


def _replace_newlines(input_str, wrapped):
    current_index = 0
    for newline_match in list(NEWLINE_REGEX.finditer(input_str)):
        (wrapped, current_index) = _replace_newline(wrapped, current_index, newline_match)
    return wrapped


def _replace_newline(wrapped, current_index, newline_match):
    newline_index = newline_match.start()
    for num, s in enumerate(wrapped, start=1):
        current_index += len(s)
        if current_index < newline_index:
            if num != len(wrapped):
                continue
            wrapped[num - 1] = f"{s}{newline_match.group()}"
            break
        wrapped[num - 1] = f"{s[:newline_index]}{newline_match.group()}{s[newline_index:]}"
        break
    return (wrapped, current_index)


def try_parse_int(input_str):
    try:
        return int(input_str)
    except ValueError:
        return None


def parse_date(input_str):
    if not input_str:
        raise ValueError("Input string was empty or None")
    if len(input_str) != 8:
        raise ValueError(f"String is not in the expected YYYYMMDD format! (len({input_str}) != 8)")
    year_str = input_str[0:4]
    month_str = input_str[4:6]
    day_str = input_str[6:8]
    try:
        year = int(year_str)
        month = int(month_str)
        day = int(day_str)
        return datetime(year, month, day)
    except Exception as e:
        raise ValueError(f"Failed to parse date from input_str ({input_str}):\n{repr(e)}")


def get_brooks_team_id(bbref_team_id):
    return BR_BB_TEAM_ID_MAP.get(bbref_team_id, bbref_team_id)


def get_bbref_team_id(brooks_team_id):
    return BB_BR_TEAM_ID_MAP.get(brooks_team_id, brooks_team_id)


def parse_timestamp(input):
    if string_is_null_or_blank(input):
        return {"hour": 0, "minute": 0}
    match = TIMESTAMP_REGEX.search(input)
    if not match:
        return {"hour": 0, "minute": 0}
    time_dict = match.groupdict()
    return {"hour": int(time_dict["hour"]), "minute": int(time_dict["minute"])}


def string_is_null_or_blank(s):
    """Check if a string is null or consists entirely of whitespace."""
    return not s or s.isspace()


def get_bbref_game_id_from_url(url):
    match = BBREF_BOXSCORE_URL_REGEX.search(url)
    return match.groupdict()["game_id"] if match else None


def get_game_date_from_bbref_game_id(bbref_game_id):
    game_date = validate_bbref_game_id(bbref_game_id).value["game_date"]
    return datetime(game_date.year, game_date.month, game_date.day)


def get_date_status_id_from_game_date(game_date):
    return game_date.strftime(DATE_ONLY_TABLE_ID)


def validate_bbref_game_id(input_str):
    match = BBREF_GAME_ID_REGEX.search(input_str)
    if not match:
        raise ValueError(f"String is not a valid bbref game id: {input_str}")
    groups = match.groupdict()
    (year, month, day, game_number) = _parse_ints_from_regex_groups(groups)
    try:
        game_date = datetime(year, month, day).date()
    except Exception as e:
        error = f"Failed to parse game_date from game_id:\n{repr(e)}"
        return Result.Fail(error)
    game_dict = {
        "game_id": match[0],
        "game_date": game_date,
        "home_team_id": groups["home_team"],
        "game_number": game_number,
    }
    return Result.Ok(game_dict)


def _parse_ints_from_regex_groups(match):
    year = int(match["year"])
    month = int(match["month"])
    day = int(match["day"])
    parsed = int(match["game_num"])
    game_number = 1 if parsed < 2 else parsed
    return (year, month, day, game_number)


def validate_bbref_game_id_list(game_ids):
    return [validate_bbref_game_id(gid).value for gid in game_ids if validate_bbref_game_id(gid).success]


def validate_brooks_game_id(input_str):
    match = BB_GAME_ID_REGEX.search(input_str)
    if not match:
        raise ValueError(f"String is not a valid bb game id: {input_str}")
    captured = match.groupdict()
    year = int(captured["year"])
    month = int(captured["month"])
    day = int(captured["day"])
    game_number = int(captured["game_num"])

    try:
        game_date = datetime(year, month, day)
    except Exception as e:
        error = f"Failed to parse game_date from game_id:\n{repr(e)}"
        return Result.Fail(error)

    away_team_id = captured["home_team"].upper()
    home_team_id = captured["away_team"].upper()

    game_dict = {
        "game_id": input_str,
        "game_date": game_date,
        "away_team_id": away_team_id,
        "home_team_id": home_team_id,
        "game_number": game_number,
    }
    return Result.Ok(game_dict)


def validate_pitch_app_id(pitch_app_id):
    match = PITCH_APP_REGEX.search(pitch_app_id)
    if not match:
        return Result.Fail(f"pitch_app_id: {pitch_app_id} is invalid")
    pitch_app_id = match[0]
    captured = match.groupdict()
    result = validate_bbref_game_id(pitch_app_id)
    game_dict = result.value
    game_dict["pitch_app_id"] = pitch_app_id
    game_dict["pitcher_id"] = captured["mlb_id"]
    pitch_app_dict = {
        "pitch_app_id": pitch_app_id,
        "game_id": game_dict["game_id"],
        "game_date": game_dict["game_date"],
        "home_team_id": game_dict["home_team_id"],
        "game_number": game_dict["game_number"],
        "pitcher_id": captured["mlb_id"],
    }
    return Result.Ok(pitch_app_dict)


def validate_at_bat_id(at_bat_id):
    match = AT_BAT_ID_REGEX.search(at_bat_id)
    if not match:
        return Result.Fail(f"at_bat_id: {at_bat_id} is invalid")
    at_bat_id = match[0]
    captured = match.groupdict()
    result = validate_bbref_game_id(at_bat_id)
    game_dict = result.value
    away_team_id = (
        captured["batter_team"] if game_dict["home_team_id"] == captured["pitcher_team"] else captured["pitcher_team"]
    )
    inning_half = "t" if game_dict["home_team_id"] == captured["pitcher_team"] else "b"
    inning_label = f"{inning_half}{captured['inning']}"
    which_half = "TOP" if inning_label[0] == "t" else "BOT"
    inning = int(captured["inning"])
    inning_str = f"{0}{inning}" if inning < 10 else inning
    inning_id = f'{game_dict["game_id"]}_INN_{which_half}{inning_str}'

    at_bat_dict = {
        "at_bat_id": at_bat_id,
        "pitch_app_id": f"{game_dict['game_id']}_{captured['pitcher_mlb_id']}",
        "game_id": game_dict["game_id"],
        "game_date": game_dict["game_date"],
        "game_number": game_dict["game_number"],
        "away_team_id": away_team_id,
        "home_team_id": game_dict["home_team_id"],
        "inning_id": inning_id,
        "inning_label": inning_label,
        "inning": captured["inning"],
        "pitcher_team": captured["pitcher_team"],
        "pitcher_mlb_id": captured["pitcher_mlb_id"],
        "batter_team": captured["batter_team"],
        "batter_mlb_id": captured["batter_mlb_id"],
        "at_bat_num": captured["at_bat_num"],
    }
    return Result.Ok(at_bat_dict)


def get_inning_id_from_at_bat_id(at_bat_id):
    at_bat_dict = validate_at_bat_id(at_bat_id).value
    inning_label = at_bat_dict["inning_label"]
    which_half = "TOP" if inning_label[0] == "t" else "BOT"
    inning = int(at_bat_dict["inning"])
    inning_str = f"{0}{inning}" if inning < 10 else inning
    return f'{at_bat_dict["game_id"]}_INN_{which_half}{inning_str}'


def inning_number_to_string(inning_str):
    if not (match := INNING_LABEL_REGEX.search(inning_str)):
        return Result.Fail(f"inning_str: {inning_str} is invalid")
    inn_half = match[1]
    inn_number = match[2]
    inn_half_str = "top" if inn_half == "TOP" else "bottom"
    inning_num_map = {
        "01": "first",
        "02": "second",
        "03": "third",
        "04": "fourth",
        "05": "fifth",
        "06": "sixth",
        "07": "seventh",
        "08": "eighth",
        "09": "ninth",
    }
    inn_num_str = inning_num_map.get(inn_number, f"{inn_number}th")
    return f"{inn_half_str} of the {inn_num_str} inning"


def replace_char_with_newlines(input_str, replace):
    if replace not in input_str:
        return input_str
    return "\n".join(s.strip() for s in input_str.split(replace))


def format_decimal_bat_stat(input):
    stat_with_leading_zero = f"{input:0.3f}"
    return stat_with_leading_zero[1:] if float(stat_with_leading_zero) < 1 else stat_with_leading_zero



"""Menu item that is used to change the current setting for an enum value."""
import subprocess

from bullet import Bullet, colors, Input, Numbers, SlidePrompt
from getch import pause

from vigorish.cli.components import print_message, yes_no_cancel_prompt, yes_no_prompt
from vigorish.cli.menu_item import MenuItem
from vigorish.config.config_setting import same_value_for_all_data_sets_is_required
from vigorish.constants import EMOJIS, MENU_NUMBERS
from vigorish.enums import ConfigType, DataSet
from vigorish.util.result import Result

NUMERIC_SETTING_UNITS = {
    "URL_SCRAPE_DELAY": "seconds",
    "BATCH_JOB_SETTINGS": "urls",
    "BATCH_SCRAPE_DELAY": "minutes",
}


class ChangeConfigSettting(MenuItem):
    def __init__(self, app, setting_name):
        super().__init__(app)
        self.menu_item_text = "Change Setting"
        self.menu_item_emoji = EMOJIS.get("SPIRAL", "")
        self.pointer = EMOJIS.get("HAND_POINTER", "")
        self.setting = self.config.all_settings.get(setting_name)
        self.setting_name_title = self.setting.setting_name_title
        self.setting_name = self.setting.setting_name
        self.data_type = self.setting.data_type
        self.possible_values = self.setting.possible_values
        self.exit_menu = False

    @property
    def enum_dict(self):
        return {
            f"{MENU_NUMBERS.get(num)}  {choice.name}": choice
            for num, choice in enumerate(self.possible_values, start=1)
        }

    @property
    def setting_units(self):
        return NUMERIC_SETTING_UNITS.get(self.setting_name, "")

    def launch(self):
        subprocess.run(["clear"])
        data_sets = self.get_data_sets_setting_will_apply_to()
        if not data_sets:
            return Result.Ok(self.exit_menu)
        setting_changed = False
        while not setting_changed:
            subprocess.run(["clear"])
            user_selections = self.get_new_setting(data_sets)
            updated_settings = self.get_updated_settings(user_selections)
            for data_set, new_value in updated_settings:
                result = self.config.change_setting(self.setting_name, data_set, new_value)
                if result.failure:
                    if "URL delay" in result.error:
                        print_message(result.error, fg="bright_red", bold=True)
                        pause(message="Press any key to continue...")
                        continue
                    return result
            setting_changed = True
        return Result.Ok(self.exit_menu)

    def get_data_sets_setting_will_apply_to(self):
        if same_value_for_all_data_sets_is_required(self.setting_name):
            return [DataSet.ALL]
        result = yes_no_cancel_prompt("Use same setting for all data sets?")
        if result.failure:
            return None
        use_same_setting = result.value
        return [DataSet.ALL] if use_same_setting else list(DataSet)

    def get_new_setting(self, data_sets):
        if self.data_type == ConfigType.NUMERIC:
            return [self.get_numeric_menu(data_set) for data_set in data_sets]
        setting_prompts = SlidePrompt(self.get_menu_prompts(data_sets))
        return setting_prompts.launch()

    def get_menu_prompts(self, data_sets):
        if self.data_type == ConfigType.ENUM:
            return [self.get_enum_menu(data_set) for data_set in data_sets]
        return [self.get_str_menu(data_set) for data_set in data_sets]

    def get_enum_menu(self, data_set):
        return Bullet(
            f"Select a value for {self.setting_name_title} (Data Set = {data_set.name}): ",
            choices=list(self.enum_dict.keys()),
            bullet="",
            shift=1,
            indent=0,
            margin=2,
            bullet_color=colors.foreground["default"],
            background_color=colors.foreground["default"],
            background_on_switch=colors.foreground["default"],
            word_color=colors.foreground["default"],
            word_on_switch=colors.bright(colors.foreground["cyan"]),
        )

    def get_str_menu(self, data_set):
        return Input(
            f"Enter a value for {self.setting_name_title} (Data Set = {data_set.name}): ",
            word_color=colors.foreground["default"],
        )

    def get_numeric_menu(self, data_set):
        if not self.setting.cannot_be_disabled:
            prompt = f"\nEnable {self.setting_name_title} (Data Set = {data_set.name})? "
            if not yes_no_prompt(prompt):
                return (prompt, (True, None, None, None, None))
        if yes_no_prompt(f"\nUse random values (Data Set = {data_set.name})? "):
            prompt, random_min, random_max, min_max_are_valid = None, 0, 0, False
            while not min_max_are_valid:
                subprocess.run(["clear"])
                prompt_min = (
                    f"Enter the minimum value (in {self.setting_units}) for "
                    f"{self.setting_name_title} (Data Set = {data_set.name}): "
                )
                prompt_max = (
                    f"Enter the maximum value (in {self.setting_units}) for "
                    f"{self.setting_name_title} (Data Set = {data_set.name}): "
                )
                random_min_prompt = Numbers(prompt_min, word_color=colors.foreground["default"])
                random_max_prompt = Numbers(prompt_max, word_color=colors.foreground["default"])
                random_min = random_min_prompt.launch()
                random_max = random_max_prompt.launch()
                if random_max > random_min:
                    min_max_are_valid = True
                    continue
                error = f"Error: maximum value ({random_max}) must be greater than minimum value ({random_min})"
                print_message(error, fg="bright_red", bold=True)
                pause(message="Press any key to continue...")
            return (prompt, (True, True, None, int(random_min), int(random_max)))
        else:
            prompt = (
                f"Enter the value (in {self.setting_units}) for {self.setting_name_title} "
                f"(Data Set = {data_set.name}): "
            )
            new_value_prompt = Numbers(prompt, word_color=colors.foreground["default"])
            new_value = new_value_prompt.launch()
            return (prompt, (True, False, int(new_value), None, None))

    def get_updated_settings(self, prompt_results):
        if self.data_type == ConfigType.ENUM:
            return self.get_updated_settings_enum(prompt_results)
        return self.get_updated_settings_str_num(prompt_results)

    def get_updated_settings_enum(self, prompt_results):
        updated_settings = []
        for prompt, selected_menu_item_text in prompt_results:
            if "ALL" in prompt:
                new_value = self.enum_dict.get(selected_menu_item_text)
                updated_settings.append((DataSet.ALL, new_value))
            for data_set in DataSet:
                if data_set.name in prompt:
                    new_value = self.enum_dict.get(selected_menu_item_text)
                    updated_settings.append((data_set, new_value))
                    break
        return updated_settings

    def get_updated_settings_str_num(self, prompt_results):
        updated_settings = []
        for prompt, new_value in prompt_results:
            if "ALL" in prompt:
                updated_settings.append((DataSet.ALL, new_value))
            for data_set in DataSet:
                if data_set.name in prompt:
                    updated_settings.append((data_set, new_value))
                    break
        return updated_settings



"""Functions for reading and writing files."""
from vigorish.enums import DataSet, LocalFileTask, S3FileTask, VigFile
from vigorish.util.result import Result
from vigorish.util.string_helpers import validate_bbref_game_id, validate_pitch_app_id


class HtmlStorage:
    """Perform CRUD operations on HTML files stored locally and/or in S3."""

    def __init__(self, config, file_helper):
        self.config = config
        self.file_helper = file_helper

    def save_html(self, data_set, url_id, html):
        local_filepath = None
        s3_object_key = None
        result_local = Result.Ok()
        result_s3 = Result.Ok()
        if self.html_stored_local(data_set):
            result_local = self.save_html_local(data_set, url_id, html)
            if result_local.success:
                local_filepath = result_local.value
        if self.html_stored_s3(data_set):  # pragma: no cover
            result_s3 = self.save_html_s3(data_set, url_id, html)
            if result_s3.success:
                s3_object_key = result_s3.value
        result = Result.Combine([result_local, result_s3])
        if result.failure:
            return result
        return Result.Ok({"local_filepath": local_filepath, "s3_object_key": s3_object_key})

    def html_stored_local(self, data_set):
        return self.file_helper.check_file_stored_local(VigFile.SCRAPED_HTML, data_set)

    def save_html_local(self, data_set, url_id, html):
        save_html_local_dict = {
            DataSet.BROOKS_GAMES_FOR_DATE: self.save_html_brooks_games_for_date_local_file,
            DataSet.BROOKS_PITCH_LOGS: self.save_html_brooks_pitch_log_local_file,
            DataSet.BROOKS_PITCHFX: self.save_html_brooks_pitchfx_local_file,
            DataSet.BBREF_GAMES_FOR_DATE: self.save_html_bbref_games_for_date_local_file,
            DataSet.BBREF_BOXSCORES: self.save_html_bbref_boxscore_local_file,
        }
        return save_html_local_dict[data_set](url_id, html)

    def html_stored_s3(self, data_set):  # pragma: no cover
        return self.file_helper.check_file_stored_s3(VigFile.SCRAPED_HTML, data_set)

    def save_html_s3(self, data_set, url_id, html):  # pragma: no cover
        save_html_s3_dict = {
            DataSet.BROOKS_GAMES_FOR_DATE: self.upload_html_brooks_games_for_date,
            DataSet.BROOKS_PITCH_LOGS: self.upload_html_brooks_pitch_log,
            DataSet.BROOKS_PITCHFX: self.upload_html_brooks_pitchfx,
            DataSet.BBREF_GAMES_FOR_DATE: self.upload_html_bbref_games_for_date,
            DataSet.BBREF_BOXSCORES: self.upload_html_bbref_boxscore,
        }
        return save_html_s3_dict[data_set](url_id, html)

    def get_html(self, data_set, url_id):
        if self.html_stored_local(data_set):
            result_local = self.get_html_local(data_set, url_id)
            if result_local.success:
                return result_local.value
        if self.html_stored_s3(data_set):  # pragma: no cover
            result_s3 = self.get_html_s3(data_set, url_id)
            if result_s3.success:
                return result_s3.value
        return None

    def get_html_local(self, data_set, url_id):
        get_html_local_dict = {
            DataSet.BROOKS_GAMES_FOR_DATE: self.get_html_brooks_games_for_date_local_file,
            DataSet.BROOKS_PITCH_LOGS: self.get_html_brooks_pitch_log_local_file,
            DataSet.BROOKS_PITCHFX: self.get_html_brooks_pitchfx_local_file,
            DataSet.BBREF_GAMES_FOR_DATE: self.get_html_bbref_games_for_date_local_file,
            DataSet.BBREF_BOXSCORES: self.get_html_bbref_boxscore_local_file,
        }
        return get_html_local_dict[data_set](url_id)

    def get_html_s3(self, data_set, url_id):  # pragma: no cover
        get_html_s3_dict = {
            DataSet.BROOKS_GAMES_FOR_DATE: self.download_html_brooks_games_for_date,
            DataSet.BROOKS_PITCH_LOGS: self.download_html_brooks_pitch_log_page,
            DataSet.BROOKS_PITCHFX: self.download_html_brooks_pitchfx_log,
            DataSet.BBREF_GAMES_FOR_DATE: self.download_html_bbref_games_for_date,
            DataSet.BBREF_BOXSCORES: self.download_html_bbref_boxscore,
        }
        return get_html_s3_dict[data_set](url_id)

    def delete_html(self, data_set, url_id):
        result_local = Result.Ok()
        result_s3 = Result.Ok()
        if self.html_stored_local(data_set):
            result_local = self.delete_html_local(data_set, url_id)
        if self.html_stored_s3(data_set):  # pragma: no cover
            result_s3 = self.delete_html_s3(data_set, url_id)
        return Result.Combine([result_local, result_s3])

    def delete_html_local(self, data_set, url_id):
        delete_html_local_dict = {
            DataSet.BROOKS_GAMES_FOR_DATE: self.delete_html_brooks_games_for_date_local_file,
            DataSet.BROOKS_PITCH_LOGS: self.delete_html_brooks_pitch_logs_local_file,
            DataSet.BROOKS_PITCHFX: self.delete_html_brooks_pitchfx_log_local_file,
            DataSet.BBREF_GAMES_FOR_DATE: self.delete_html_bbref_games_for_date_local_file,
            DataSet.BBREF_BOXSCORES: self.delete_html_bbref_boxscore_local_file,
        }
        return delete_html_local_dict[data_set](url_id)

    def delete_html_s3(self, data_set, url_id):  # pragma: no cover
        delete_html_s3_dict = {
            DataSet.BROOKS_GAMES_FOR_DATE: self.delete_html_brooks_games_for_date_s3,
            DataSet.BROOKS_PITCH_LOGS: self.delete_html_brooks_pitch_logs_s3,
            DataSet.BROOKS_PITCHFX: self.delete_html_brooks_pitchfx_log_s3,
            DataSet.BBREF_GAMES_FOR_DATE: self.delete_html_bbref_games_for_date_s3,
            DataSet.BBREF_BOXSCORES: self.delete_html_bbref_boxscore_s3,
        }
        return delete_html_s3_dict[data_set](url_id)

    def save_html_brooks_games_for_date_local_file(self, game_date, html):
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.WRITE_FILE,
            data_set=DataSet.BROOKS_GAMES_FOR_DATE,
            file_type=VigFile.SCRAPED_HTML,
            game_date=game_date,
            scraped_data=html,
        )

    def save_html_brooks_pitch_log_local_file(self, pitch_app_id, html):
        result = validate_pitch_app_id(pitch_app_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.WRITE_FILE,
            data_set=DataSet.BROOKS_PITCH_LOGS,
            file_type=VigFile.SCRAPED_HTML,
            game_date=game_dict["game_date"],
            scraped_data=html,
            pitch_app_id=pitch_app_id,
        )

    def save_html_brooks_pitchfx_local_file(self, pitch_app_id, html):
        result = validate_pitch_app_id(pitch_app_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.WRITE_FILE,
            data_set=DataSet.BROOKS_PITCHFX,
            file_type=VigFile.SCRAPED_HTML,
            game_date=game_dict["game_date"],
            scraped_data=html,
            pitch_app_id=pitch_app_id,
        )

    def save_html_bbref_games_for_date_local_file(self, game_date, html):
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.WRITE_FILE,
            data_set=DataSet.BBREF_GAMES_FOR_DATE,
            file_type=VigFile.SCRAPED_HTML,
            game_date=game_date,
            scraped_data=html,
        )

    def save_html_bbref_boxscore_local_file(self, bbref_game_id, html):
        result = validate_bbref_game_id(bbref_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.WRITE_FILE,
            data_set=DataSet.BBREF_BOXSCORES,
            file_type=VigFile.SCRAPED_HTML,
            game_date=game_dict["game_date"],
            scraped_data=html,
            bbref_game_id=bbref_game_id,
        )

    def get_html_brooks_games_for_date_local_file(self, game_date):
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.READ_FILE,
            data_set=DataSet.BROOKS_GAMES_FOR_DATE,
            file_type=VigFile.SCRAPED_HTML,
            game_date=game_date,
        )

    def get_html_brooks_pitch_log_local_file(self, pitch_app_id):
        result = validate_pitch_app_id(pitch_app_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.READ_FILE,
            data_set=DataSet.BROOKS_PITCH_LOGS,
            file_type=VigFile.SCRAPED_HTML,
            game_date=game_dict["game_date"],
            pitch_app_id=pitch_app_id,
        )

    def get_html_brooks_pitchfx_local_file(self, pitch_app_id):
        result = validate_pitch_app_id(pitch_app_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.READ_FILE,
            data_set=DataSet.BROOKS_PITCHFX,
            file_type=VigFile.SCRAPED_HTML,
            game_date=game_dict["game_date"],
            pitch_app_id=pitch_app_id,
        )

    def get_html_bbref_games_for_date_local_file(self, game_date):
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.READ_FILE,
            data_set=DataSet.BBREF_GAMES_FOR_DATE,
            file_type=VigFile.SCRAPED_HTML,
            game_date=game_date,
        )

    def get_html_bbref_boxscore_local_file(self, bbref_game_id):
        result = validate_bbref_game_id(bbref_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.READ_FILE,
            data_set=DataSet.BBREF_BOXSCORES,
            file_type=VigFile.SCRAPED_HTML,
            game_date=game_dict["game_date"],
            bbref_game_id=bbref_game_id,
        )

    def upload_html_brooks_games_for_date(self, game_date, html):  # pragma: no cover
        result = self.save_html_local(DataSet.BROOKS_GAMES_FOR_DATE, game_date, html)
        if result.failure:
            return result
        return self.file_helper.perform_s3_task(
            task=S3FileTask.UPLOAD,
            file_type=VigFile.SCRAPED_HTML,
            data_set=DataSet.BROOKS_GAMES_FOR_DATE,
            game_date=game_date,
        )

    def upload_html_brooks_pitch_log(self, pitch_app_id, html):  # pragma: no cover
        result = self.save_html_local(DataSet.BROOKS_PITCH_LOGS, pitch_app_id, html)
        if result.failure:
            return result
        result = validate_pitch_app_id(pitch_app_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_s3_task(
            task=S3FileTask.UPLOAD,
            file_type=VigFile.SCRAPED_HTML,
            data_set=DataSet.BROOKS_PITCH_LOGS,
            game_date=game_dict["game_date"],
            pitch_app_id=pitch_app_id,
        )

    def upload_html_brooks_pitchfx(self, pitch_app_id, html):  # pragma: no cover
        result = self.save_html_local(DataSet.BROOKS_PITCHFX, pitch_app_id, html)
        if result.failure:
            return result
        result = validate_pitch_app_id(pitch_app_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_s3_task(
            task=S3FileTask.UPLOAD,
            file_type=VigFile.SCRAPED_HTML,
            data_set=DataSet.BROOKS_PITCHFX,
            game_date=game_dict["game_date"],
            pitch_app_id=pitch_app_id,
        )

    def upload_html_bbref_games_for_date(self, game_date, html):  # pragma: no cover
        result = self.save_html_local(DataSet.BBREF_GAMES_FOR_DATE, game_date, html)
        if result.failure:
            return result
        return self.file_helper.perform_s3_task(
            task=S3FileTask.UPLOAD,
            file_type=VigFile.SCRAPED_HTML,
            data_set=DataSet.BBREF_GAMES_FOR_DATE,
            game_date=game_date,
        )

    def upload_html_bbref_boxscore(self, bbref_game_id, html):  # pragma: no cover
        result = self.save_html_local(DataSet.BBREF_BOXSCORES, bbref_game_id, html)
        if result.failure:
            return result
        result = validate_bbref_game_id(bbref_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_s3_task(
            task=S3FileTask.UPLOAD,
            file_type=VigFile.SCRAPED_HTML,
            data_set=DataSet.BBREF_BOXSCORES,
            game_date=game_dict["game_date"],
            bbref_game_id=bbref_game_id,
        )

    def download_html_brooks_games_for_date(self, game_date):  # pragma: no cover
        """Download raw HTML for brooks daily scoreboard page."""
        return self.file_helper.perform_s3_task(
            task=S3FileTask.DOWNLOAD,
            data_set=DataSet.BROOKS_GAMES_FOR_DATE,
            file_type=VigFile.SCRAPED_HTML,
            game_date=game_date,
        )

    def download_html_brooks_pitch_log_page(self, pitch_app_id):  # pragma: no cover
        """Download raw HTML for brooks pitch log page for a single pitching appearance."""
        result = validate_pitch_app_id(pitch_app_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_s3_task(
            task=S3FileTask.DOWNLOAD,
            data_set=DataSet.BROOKS_PITCH_LOGS,
            file_type=VigFile.SCRAPED_HTML,
            game_date=game_dict["game_date"],
            pitch_app_id=pitch_app_id,
        )

    def download_html_brooks_pitchfx_log(self, pitch_app_id):  # pragma: no cover
        """Download raw HTML for brooks pitchfx data for a single pitching appearance."""
        result = validate_pitch_app_id(pitch_app_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_s3_task(
            task=S3FileTask.DOWNLOAD,
            file_type=VigFile.SCRAPED_HTML,
            data_set=DataSet.BROOKS_PITCHFX,
            game_date=game_dict["game_date"],
            pitch_app_id=pitch_app_id,
        )

    def download_html_bbref_games_for_date(self, game_date):  # pragma: no cover
        """Download raw HTML for bbref daily scoreboard page."""
        return self.file_helper.perform_s3_task(
            task=S3FileTask.DOWNLOAD,
            data_set=DataSet.BBREF_GAMES_FOR_DATE,
            file_type=VigFile.SCRAPED_HTML,
            game_date=game_date,
        )

    def download_html_bbref_boxscore(self, bbref_game_id):  # pragma: no cover
        """Download raw HTML for bbref daily scoreboard page."""
        result = validate_bbref_game_id(bbref_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_s3_task(
            task=S3FileTask.DOWNLOAD,
            file_type=VigFile.SCRAPED_HTML,
            data_set=DataSet.BBREF_BOXSCORES,
            game_date=game_dict["game_date"],
            bbref_game_id=bbref_game_id,
        )

    def delete_html_brooks_games_for_date_local_file(self, game_date):
        """Delete scraped HTML for brooks daily scoreboard page from local folder."""
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.DELETE_FILE,
            data_set=DataSet.BROOKS_GAMES_FOR_DATE,
            file_type=VigFile.SCRAPED_HTML,
            game_date=game_date,
        )

    def delete_html_brooks_games_for_date_s3(self, game_date):  # pragma: no cover
        """Delete scraped HTML for brooks daily scoreboard page from S3."""
        return self.file_helper.perform_s3_task(
            task=S3FileTask.DELETE,
            data_set=DataSet.BROOKS_GAMES_FOR_DATE,
            file_type=VigFile.SCRAPED_HTML,
            game_date=game_date,
        )

    def delete_html_bbref_games_for_date_local_file(self, game_date):
        """Delete scraped HTML for bbref daily scoreboard page from local folder."""
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.DELETE_FILE,
            data_set=DataSet.BBREF_GAMES_FOR_DATE,
            file_type=VigFile.SCRAPED_HTML,
            game_date=game_date,
        )

    def delete_html_bbref_games_for_date_s3(self, game_date):  # pragma: no cover
        """Delete scraped HTML for bbref daily scoreboard page from S3."""
        return self.file_helper.perform_s3_task(
            task=S3FileTask.DELETE,
            data_set=DataSet.BBREF_GAMES_FOR_DATE,
            file_type=VigFile.SCRAPED_HTML,
            game_date=game_date,
        )

    def delete_html_bbref_boxscore_local_file(self, bbref_game_id):
        """Delete scraped HTML for bbref boxscore page from local folder."""
        result = validate_bbref_game_id(bbref_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.DELETE_FILE,
            data_set=DataSet.BBREF_BOXSCORES,
            file_type=VigFile.SCRAPED_HTML,
            game_date=game_dict["game_date"],
            bbref_game_id=bbref_game_id,
        )

    def delete_html_bbref_boxscore_s3(self, bbref_game_id):  # pragma: no cover
        """Delete scraped HTML for bbref boxscore page from S3."""
        result = validate_bbref_game_id(bbref_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_s3_task(
            task=S3FileTask.DELETE,
            data_set=DataSet.BBREF_BOXSCORES,
            file_type=VigFile.SCRAPED_HTML,
            game_date=game_dict["game_date"],
            bbref_game_id=bbref_game_id,
        )

    def delete_html_brooks_pitch_logs_local_file(self, pitch_app_id):
        """Delete scraped HTML for brooks pitch log page from local folder."""
        result = validate_pitch_app_id(pitch_app_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.DELETE_FILE,
            data_set=DataSet.BROOKS_PITCH_LOGS,
            file_type=VigFile.SCRAPED_HTML,
            game_date=game_dict["game_date"],
            pitch_app_id=pitch_app_id,
        )

    def delete_html_brooks_pitch_logs_s3(self, pitch_app_id):  # pragma: no cover
        """Delete scraped HTML for brooks pitch log page from S3."""
        result = validate_pitch_app_id(pitch_app_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_s3_task(
            task=S3FileTask.DELETE,
            data_set=DataSet.BROOKS_PITCH_LOGS,
            file_type=VigFile.SCRAPED_HTML,
            game_date=game_dict["game_date"],
            pitch_app_id=pitch_app_id,
        )

    def delete_html_brooks_pitchfx_log_local_file(self, pitch_app_id):
        result = validate_pitch_app_id(pitch_app_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.DELETE_FILE,
            data_set=DataSet.BROOKS_PITCHFX,
            file_type=VigFile.SCRAPED_HTML,
            game_date=game_dict["game_date"],
            pitch_app_id=pitch_app_id,
        )

    def delete_html_brooks_pitchfx_log_s3(self, pitch_app_id):  # pragma: no cover
        """Delete scraped HTML for brooks pitchfx data from s3."""
        result = validate_pitch_app_id(pitch_app_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_s3_task(
            task=S3FileTask.DELETE,
            data_set=DataSet.BROOKS_PITCHFX,
            file_type=VigFile.SCRAPED_HTML,
            game_date=game_dict["game_date"],
            pitch_app_id=pitch_app_id,
        )



from dataclasses import dataclass, field
from typing import List

from vigorish.scrape.bbref_boxscores.models.bat_stats_detail import BBRefBatStatsDetail
from vigorish.util.list_helpers import as_dict_list


@dataclass
class BBRefBatStats:
    """Statistics for all plate appearances that occurred for one player in a single game."""

    player_id_br: str = ""
    player_team_id_br: str = ""
    opponent_team_id_br: str = ""
    plate_appearances: str = "0"
    at_bats: str = "0"
    hits: str = "0"
    runs_scored: str = "0"
    rbis: str = "0"
    bases_on_balls: str = "0"
    strikeouts: str = "0"
    avg_to_date: str = "0"
    obp_to_date: str = "0"
    slg_to_date: str = "0"
    ops_to_date: str = "0"
    total_pitches: str = "0"
    total_strikes: str = "0"
    wpa_bat: str = "0"
    avg_lvg_index: str = "0"
    wpa_bat_pos: str = "0"
    wpa_bat_neg: str = "0"
    re24_bat: str = "0"
    details: List[BBRefBatStatsDetail] = field(default_factory=list)

    def as_dict(self):
        """Convert batting statistics for one player in a single game to a dictionary."""
        return {
            "player_id_br": self.player_id_br,
            "player_team_id_br": self.player_team_id_br,
            "opponent_team_id_br": self.opponent_team_id_br,
            "plate_appearances": int(self.plate_appearances),
            "at_bats": int(self.at_bats),
            "hits": int(self.hits),
            "runs_scored": int(self.runs_scored),
            "rbis": int(self.rbis),
            "bases_on_balls": int(self.bases_on_balls),
            "strikeouts": int(self.strikeouts),
            "avg_to_date": float(self.avg_to_date),
            "obp_to_date": float(self.obp_to_date),
            "slg_to_date": float(self.slg_to_date),
            "ops_to_date": float(self.ops_to_date),
            "total_pitches": int(self.total_pitches),
            "total_strikes": int(self.total_strikes),
            "wpa_bat": float(self.wpa_bat),
            "avg_lvg_index": float(self.avg_lvg_index),
            "wpa_bat_pos": float(self.wpa_bat_pos),
            "wpa_bat_neg": float(self.wpa_bat_neg),
            "re24_bat": float(self.re24_bat),
            "details": as_dict_list(self.details),
        }



from sqlalchemy import func, join, select
from sqlalchemy_utils import create_view

import vigorish.database as db


class Season_Date_View(db.Base):
    __table__ = create_view(
        name="season_date_status",
        selectable=select(
            [
                db.Season.id.label("id"),
                func.count(db.DateScrapeStatus.id).label("total_days"),
                func.sum(db.DateScrapeStatus.scraped_daily_dash_bbref).label("total_scraped_daily_dash_bbref"),
                func.sum(db.DateScrapeStatus.scraped_daily_dash_brooks).label("total_scraped_daily_dash_brooks"),
                func.sum(db.DateScrapeStatus.game_count_bbref).label("total_game_count_bbref"),
                func.sum(db.DateScrapeStatus.game_count_brooks).label("total_game_count_brooks"),
            ]
        )
        .select_from(
            join(
                db.Season,
                db.DateScrapeStatus,
                db.Season.id == db.DateScrapeStatus.season_id,
            )
        )
        .group_by(db.Season.id),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )



from dataclasses import dataclass, field
from datetime import datetime

import vigorish.database as db
from vigorish.enums import DataSet
from vigorish.patch.base import Patch, PatchList
from vigorish.util.dt_format_strings import DATE_ONLY
from vigorish.util.result import Result


@dataclass
class BrooksGamesForDatePatchList(PatchList):
    @property
    def game_date(self):
        return datetime.strptime(self.url_id, DATE_ONLY) if self.url_id else None

    def __post_init__(self):
        self.data_set = DataSet.BROOKS_GAMES_FOR_DATE
        self.patch_list_id = "__brooks_games_for_date_patch_list__"

    def apply(self, data, db_session):
        for patch in self.patch_list:
            data = patch.apply(data)
            patch.apply_to_database(data, db_session)
        return Result.Ok(data)


@dataclass
class PatchBrooksGamesForDateBBRefGameID(Patch):
    game_date: str = field(repr=False)
    invalid_bbref_game_id: str
    valid_bbref_game_id: str

    def __post_init__(self):
        self.data_set = DataSet.BROOKS_GAMES_FOR_DATE
        self.patch_id = "__patch_brooks_games_for_date_bbref_game_id__"

    def apply(self, data):
        if self.invalid_bbref_game_id in data.all_bbref_game_ids:
            for game_info in data.games:
                if game_info.bbref_game_id == self.invalid_bbref_game_id:
                    game_info.bbref_game_id = self.valid_bbref_game_id
                    break
        return data

    def apply_to_database(self, data, db_session):
        delete_pitchfx_with_invalid_id(db_session, self.invalid_bbref_game_id)
        delete_pitch_apps_with_invalid_id(db_session, self.invalid_bbref_game_id)
        delete_game_status(db_session, self.invalid_bbref_game_id)
        update_game_status(db_session, data, self.valid_bbref_game_id)
        db_session.commit()


@dataclass
class PatchBrooksGamesForDateRemoveGame(Patch):
    game_date: str = field(repr=False)
    remove_bbref_game_id: str

    def __post_init__(self):
        self.data_set = DataSet.BROOKS_GAMES_FOR_DATE
        self.patch_id = "__patch_brooks_games_for_date_remove_game__"

    def apply(self, data):
        data.games = list(filter(lambda x: x.bbref_game_id != self.remove_bbref_game_id, data.games))
        data.game_count = len(data.games)
        return data

    def apply_to_database(self, data, db_session):
        delete_pitchfx_with_invalid_id(db_session, self.remove_bbref_game_id)
        delete_pitch_apps_with_invalid_id(db_session, self.remove_bbref_game_id)
        delete_game_status(db_session, self.remove_bbref_game_id)
        update_date_status(db_session, data)
        db_session.commit()


def delete_pitchfx_with_invalid_id(db_session, bbref_game_id):
    all_pfx_with_invalid_id = db_session.query(db.PitchFx).filter_by(bbref_game_id=bbref_game_id).all()
    for pfx in all_pfx_with_invalid_id:
        db_session.delete(pfx)


def delete_pitch_apps_with_invalid_id(db_session, bbref_game_id):
    all_pitch_apps_with_invalid_id = (
        db_session.query(db.PitchAppScrapeStatus).filter_by(bbref_game_id=bbref_game_id).all()
    )
    for pitch_app in all_pitch_apps_with_invalid_id:
        db_session.delete(pitch_app)


def delete_game_status(db_session, bbref_game_id):
    game_status = db.GameScrapeStatus.find_by_bbref_game_id(db_session, bbref_game_id)
    if game_status:
        db_session.delete(game_status)


def update_game_status(db_session, data, bbref_game_id):
    game_status = db.GameScrapeStatus.find_by_bbref_game_id(db_session, bbref_game_id)
    if game_status:
        for game_info in data.games:
            if game_info.bbref_game_id == bbref_game_id:
                game_status.bb_game_id = game_info.bb_game_id
                game_status.game_time_hour = game_info.game_time_hour
                game_status.game_time_minute = game_info.game_time_minute
                game_status.game_time_zone = game_info.time_zone_name
                game_status.pitch_app_count_brooks = game_info.pitcher_appearance_count
                break


def update_date_status(db_session, data):
    date_status = db.DateScrapeStatus.find_by_date(db_session, data.game_date)
    date_status.game_count_brooks = data.game_count
